import time
from abc import ABC, abstractmethod
from multiprocessing import Queue
from queue import Empty
from threading import Event
from typing import Generator

import pika
from pika.adapters.blocking_connection import BlockingChannel, BlockingConnection
from pika.exceptions import AMQPConnectionError

from abstra_internals.environment import (
    EXECUTION_QUEUE_CONCURRENCY,
    RABBITMQ_CONNECTION_TIMEOUT_SECONDS,
    RABBITMQ_EXECUTION_QUEUE,
    RABBITMQ_RETRY_INITIAL_DELAY_SECONDS,
    RABBITMQ_RETRY_MAX_ATTEMPTS,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.producer import PreExecution, QueueMessage
from abstra_internals.utils import deserialize


class Consumer(ABC):
    @abstractmethod
    def threadsafe_ack(self, msg: QueueMessage):
        raise NotImplementedError

    @abstractmethod
    def threadsafe_nack(self, msg: QueueMessage):
        raise NotImplementedError

    @abstractmethod
    def iter(self) -> Generator[QueueMessage, None, None]:
        raise NotImplementedError

    @abstractmethod
    def stop_iter(self):
        raise NotImplementedError


class RabbitConsumer(Consumer):
    """All methods are not thread-safe unless specified otherwise."""

    channel: BlockingChannel
    connection: BlockingConnection
    conn_uri: str

    def __init__(self, conn_uri: str):
        self.concurrency = EXECUTION_QUEUE_CONCURRENCY
        self.queue = RABBITMQ_EXECUTION_QUEUE
        self.conn_uri = conn_uri
        self.stop_evt = Event()

        self._connect()

    def _connect(self):
        """
        Establish RabbitMQ connection with exponential backoff retry logic.
        Retries connection attempts with exponential backoff to handle
        transient network issues during pod startup.
        """
        AbstraLogger.info("[RabbitConsumer] Connecting...")

        params = pika.URLParameters(self.conn_uri)
        params.connection_attempts = 1  # Single attempt per retry iteration
        params.socket_timeout = RABBITMQ_CONNECTION_TIMEOUT_SECONDS
        params.blocked_connection_timeout = RABBITMQ_CONNECTION_TIMEOUT_SECONDS

        delay = RABBITMQ_RETRY_INITIAL_DELAY_SECONDS
        last_exception = None

        for attempt in range(1, RABBITMQ_RETRY_MAX_ATTEMPTS + 1):
            try:
                self.connection = pika.BlockingConnection(params)
                self.channel = self.connection.channel()
                self.channel.basic_qos(prefetch_count=self.concurrency)
                self.channel.queue_declare(queue=self.queue, durable=True)
                AbstraLogger.info(
                    "[RabbitConsumer] Connection established successfully"
                )
                return
            except AMQPConnectionError as e:
                last_exception = e
                if attempt < RABBITMQ_RETRY_MAX_ATTEMPTS:
                    AbstraLogger.warning(
                        f"[RabbitConsumer] Connection failed (attempt {attempt}): {e}. "
                        f"Retrying in {delay}s..."
                    )
                    time.sleep(delay)
                    delay = min(delay * 2, 30)  # Exponential backoff, max 30s
                else:
                    AbstraLogger.error(
                        f"[RabbitConsumer] All {RABBITMQ_RETRY_MAX_ATTEMPTS} connection attempts failed"
                    )

        raise last_exception or AMQPConnectionError("Failed to connect to RabbitMQ")

    def threadsafe_ack(self, msg: QueueMessage):
        AbstraLogger.debug(
            f"[RabbitConsumer] Adding ack callback for [{msg.delivery_tag}]"
        )

        def callback():
            AbstraLogger.debug(
                f"[RabbitConsumer] Acknowledging message [{msg.delivery_tag}]"
            )
            self.channel.basic_ack(delivery_tag=msg.delivery_tag)

        self.connection.add_callback_threadsafe(callback)

    def threadsafe_nack(self, msg: QueueMessage):
        AbstraLogger.debug(
            f"[RabbitConsumer] Adding nack callback for [{msg.delivery_tag}]"
        )

        def callback():
            AbstraLogger.debug(
                f"[RabbitConsumer] Not acknowledging message [{msg.delivery_tag}]"
            )
            self.channel.basic_nack(delivery_tag=msg.delivery_tag, requeue=False)

        self.connection.add_callback_threadsafe(callback)

    def iter(self) -> Generator[QueueMessage, None, None]:
        AbstraLogger.info("[RabbitConsumer] Starting consumer iteration")

        for method, _, body in self.channel.consume(
            queue=self.queue, inactivity_timeout=10, auto_ack=False
        ):
            if self.stop_evt.is_set():
                break

            if not method:
                AbstraLogger.debug("[RabbitConsumer] Consume returned no messages")
                continue

            yield QueueMessage(
                preexecution=PreExecution(**deserialize(body.decode("utf-8"))),
                delivery_tag=method.delivery_tag,
            )

    def stop_iter(self):
        AbstraLogger.warning("[RabbitConsumer] Setting stop event for consumer")
        self.stop_evt.set()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_iter()

        AbstraLogger.debug("[RabbitConsumer] Exiting consumer context manager")
        self.connection.process_data_events(time_limit=60)
        AbstraLogger.warning("[RabbitConsumer] Data events processed")

        if self.connection.is_open:
            AbstraLogger.warning("[RabbitConsumer] Cancelling channel")
            self.channel.cancel()
            AbstraLogger.warning("[RabbitConsumer] Closing channel")
            self.channel.close()
            AbstraLogger.warning("[RabbitConsumer] Closing connection")
            self.connection.close()

        return False

    def __enter__(self) -> "RabbitConsumer":
        return self


class EditorConsumer(Consumer):
    def __init__(self, queue: Queue) -> None:
        self.stop_evt = Event()
        self.queue = queue

    def threadsafe_nack(self, msg: QueueMessage):
        pass

    def threadsafe_ack(self, msg: QueueMessage):
        pass

    def stop_iter(self):
        self.stop_evt.set()

    def iter(self) -> Generator[QueueMessage, None, None]:
        while True:
            if self.stop_evt.is_set():
                break

            try:
                if msg := self.queue.get(timeout=1):
                    yield msg
            except Empty:
                continue



import json
from typing import TYPE_CHECKING, List, Optional, cast

from abstra_internals.entities.forms.widgets.widget_base import OutputWidget

if TYPE_CHECKING:
    import pandas as pd


class PandasOutput(OutputWidget):
    """Pandas DataFrame output widget for displaying tabular data.

    Attributes:
        df (pd.DataFrame): The pandas DataFrame to display.
    """

    type = "pandas-output"

    def __init__(
        self,
        df: "pd.DataFrame",
        *,
        label: Optional[str] = None,
        key: Optional[str] = None,
        actions: Optional[List[dict]] = None,
        full_width: bool = False,
        display_index: bool = True,
        filterable: bool = True,
        page_size: int = 10,
        pagination_always_visible: bool = True,
    ):
        """Initialize a PandasOutput widget.

        Args:
            df (pd.DataFrame): The pandas DataFrame to display.
            label (Optional[str]): Text label displayed above the table.
            key (Optional[str]): Identifier for the widget.
            actions (Optional[List[dict]]): List of action configurations.
            full_width (bool): Whether the table should take up the full width of its container.
            display_index (bool): Whether to display row indices.
            filterable (bool): Whether the table is filterable.
            page_size (int): Number of rows to display per page.
            pagination_always_visible (bool): Whether pagination controls are always visible.
        """
        self.label = label
        self.key = key
        self.df = df.reset_index(drop=True) if df is not None else None
        self.actions = actions
        self.full_width = full_width
        self.display_index = display_index
        self.filterable = filterable
        self.page_size = page_size
        self.pagination_always_visible = pagination_always_visible

    def serialize_table(self):
        if self.df is None:
            import pandas as pd

            df = pd.DataFrame({"change the": [1, 2, 3], "df property": [4, 5, 6]})
            df_json = df.to_json(orient="table")
            if not isinstance(df_json, str):
                raise Exception("df.to_json() did not return a string")
            serialized = json.loads(df_json)
            del serialized["schema"]["pandas_version"]
            return serialized
        serialized = json.loads(cast(str, self.df.to_json(orient="table")))
        del serialized["schema"]["pandas_version"]
        return serialized

    def _render(self):
        return {
            "type": self.type,
            "table": self.serialize_table(),
            "fullWidth": self.full_width,
            "displayIndex": self.display_index,
            "label": self.label,
            "actions": self.actions,
            "filterable": self.filterable,
            "pageSize": self.page_size,
            "paginationAlwaysVisible": self.pagination_always_visible,
        }



from abstra_internals.controllers.execution.consumer import ConsumerController
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import DEFAULT_PORT, RABBITMQ_CONNECTION_URI
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.consumer import RabbitConsumer
from abstra_internals.repositories.factory import build_prod_repositories
from abstra_internals.settings import SettingsController
from abstra_internals.signals import SignalHandlers


def run():
    SignalHandlers.init()
    AbstraLogger.init("cloud")
    SettingsController.set_root_path(".")
    SettingsController.set_server_port(DEFAULT_PORT)

    if not RABBITMQ_CONNECTION_URI:
        raise Exception("RABBITMQ_CONNECTION_URI not found")

    controller = MainController(repositories=build_prod_repositories())

    with RabbitConsumer(RABBITMQ_CONNECTION_URI) as consumer:
        SignalHandlers.register_sigterm_callback(consumer.stop_iter)
        ConsumerController(controller, consumer).start_loop()


if __name__ == "__main__":
    run()



from typing import Any, Callable

from abstra_internals.controllers.sdk.sdk_context import SDKContextStore


class ReuseCacheInterface:
    def __init__(self) -> None:
        self.controller = SDKContextStore.get_by_thread().form_sdk
        self.seq = 0

    def reuse(self, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        """
        Reuse the result of a function call with the given arguments.

        Args:
            func (Callable): The function to reuse.
            *args (Any): Variable length argument list to pass to the function.
            **kwargs (Any): Arbitrary keyword arguments to pass to the function.

        Returns:
            Any: The result of the function call.
        """
        return self.controller.reuse(func, *args, **kwargs)


def reuse(func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
    """
    Reuse the result of a function call with the given arguments.

    Args:
        func (Callable): The function to reuse.
        args (Any): Variable length argument list to pass to the function.
        kwargs (Any): Arbitrary keyword arguments to pass to the function.

    Returns:
        Any: The result of the function call.
    """
    return ReuseCacheInterface().reuse(func, *args, **kwargs)



from typing import Callable, Optional, Union

from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.template import (
    Template,
    TemplateFunction,
    TemplateGenerator,
    TemplateGeneratorFunction,
    TemplateWithButtons,
)

ComputationFunction = Callable[[State], None]


class Step:
    def run(self, state: State) -> Optional[TemplateWithButtons]:
        raise NotImplementedError()

    def normalize_result(
        self, result: Optional[Union[Template, TemplateWithButtons]]
    ) -> Optional[TemplateWithButtons]:
        if result is None:
            return None
        if isinstance(result, tuple):
            return result
        if isinstance(result, list):
            return result, None
        raise ValueError(f"Invalid result type: {type(result)}")


class PageStep(Step):
    def __init__(self, func: TemplateFunction):
        self.func = func

    def run(self, state: State) -> Optional[TemplateWithButtons]:
        return self.normalize_result(self.func(state))


class ComputationStep(Step):
    def __init__(self, func: ComputationFunction):
        self.func = func

    def run(self, state: State) -> Optional[TemplateWithButtons]:
        self.func(state)
        return None


class GeneratorStep(ComputationStep):
    generator: Optional[TemplateGenerator]

    def __init__(self, func: TemplateGeneratorFunction):
        self.func = func
        self.generator = None

    def get_generator(self, state: State):
        if self.generator is None:
            self.generator = self.func(state)
        return self.generator

    def run(self, state: State) -> Optional[TemplateWithButtons]:
        generator = self.get_generator(state)

        try:
            return self.normalize_result(next(generator))
        except StopIteration:
            self.generator = None
            return None


class EndPageStep(PageStep):
    def __init__(self, func: TemplateFunction):
        super().__init__(func)



from abstra_internals.repositories.linter.models import (
    LinterFix,
    LinterIssue,
    LinterRule,
)
from abstra_internals.repositories.project.project import (
    FormStage,
    LocalProjectRepository,
    Project,
)
from abstra_internals.utils.file import generate_conflictless_path, is_path_in_conflict


class ConflictingPathFix(LinterFix):
    label = "Fix conflicting path"
    description = "Change the path of the form to avoid conflicts"

    def __init__(
        self,
        form: FormStage,
        project: Project,
        project_repository: LocalProjectRepository,
    ):
        self.form = form
        self.project = project
        self.project_repository = project_repository

    def fix(self):
        self.project.update_stage(
            self.form, {"path": generate_conflictless_path(self.form.path)}
        )
        self.project_repository.save(self.project)


class ConflictingPathFound(LinterIssue):
    type = "bug"
    fixes = []

    def __init__(
        self,
        form: FormStage,
        project: Project,
        project_repository: LocalProjectRepository,
    ):
        self.label = f"The path of the form {form.title} is in conflict with an internal reserved path. This can cause unexpected behavior. You can either change it manually in the Editor or use the 'Fix conflicting path' button."
        self.fixes = [ConflictingPathFix(form, project, project_repository)]


class ConflictingPath(LinterRule):
    label = "Conflicting path"
    type = "bug"

    def find_issues(self):
        project_repository = LocalProjectRepository()
        project = project_repository.load()

        issues = []
        for form in project.forms:
            if is_path_in_conflict(form.path):
                issues.append(ConflictingPathFound(form, project, project_repository))

        return issues



from concurrent.futures import ThreadPoolExecutor
from uuid import uuid4

from abstra_internals.controllers.execution.worker_process import process_main
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import (
    EXECUTION_QUEUE_CONCURRENCY,
    PROCESS_TIMEOUT_SECONDS,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.consumer import Consumer, QueueMessage
from abstra_internals.repositories.producer import LocalProducerRepository
from abstra_internals.settings import Settings


class StageNotFound(Exception):
    pass


class NonCleanExit(Exception):
    pass


class ConsumerController:
    def __init__(self, main_controller: MainController, consumer: Consumer):
        self.main_controller = main_controller
        self.consumer = consumer
        self.app_id = str(uuid4())
        self.concurrency = EXECUTION_QUEUE_CONCURRENCY

    def start_loop(self):
        AbstraLogger.debug(
            f"[ConsumerController] Starting loop with {self.concurrency} threads"
        )

        thread_pool = ThreadPoolExecutor(
            max_workers=self.concurrency,
            thread_name_prefix="ExecutionConsumer",
        )

        try:
            for msg in self.consumer.iter():
                AbstraLogger.debug(
                    f"[ConsumerController] Submitting message [{msg.delivery_tag}] for processing"
                )
                thread_pool.submit(
                    self.run_subprocess,
                    msg=msg,
                )
        finally:
            AbstraLogger.warning(
                "[ConsumerController] Consumer main loop exited, waiting for threads to finish"
            )
            thread_pool.shutdown(wait=True)
            AbstraLogger.warning("[ConsumerController] All threads finished")

            # If the loop exits and there are running executions, gracefull shutdown has failed
            self.main_controller.fail_app_executions(
                app_id=self.app_id,
                reason="Failed to set status",
            )

    def run_subprocess(self, msg: QueueMessage):
        worker_id = str(uuid4())
        head_id = worker_id.split("-")[0]

        try:
            mp_context = self.main_controller.repositories.mp_context.get_context()
            stage = self.main_controller.get_stage(msg.preexecution.stage_id)

            if stage is None:
                AbstraLogger.warning(
                    f"[ConsumerController] Stage not found: {msg.preexecution.stage_id}. Message [{msg.delivery_tag}] will be acknowledged."
                )
                self.consumer.threadsafe_ack(msg)
                return

            local_queue = None
            if isinstance(
                self.main_controller.repositories.producer, LocalProducerRepository
            ):
                local_queue = self.main_controller.repositories.producer.queue

            p = mp_context.Process(
                target=process_main,
                name=f"Worker-{head_id}",
                kwargs=dict(
                    stage=stage,
                    worker_id=worker_id,
                    app_id=self.app_id,
                    root_path=Settings.root_path,
                    server_port=Settings.server_port,
                    request=msg.preexecution.context,
                    local_queue=local_queue,
                ),
            )

            p.start()
            p.join(timeout=PROCESS_TIMEOUT_SECONDS)

            # If timeout is reached, the process is still alive
            if p.is_alive():
                p.terminate()
                raise NonCleanExit(
                    f"Run took too long to complete ({PROCESS_TIMEOUT_SECONDS} secs) and was terminated. Please make sure all your requests calls have a timeout set."
                )

            if p.exitcode != 0:
                err_msg = f"Run failed with status '{p.exitcode}'"
                if p.exitcode == -9:
                    err_msg += ": memory limit reached."

                raise NonCleanExit(err_msg)

            self.consumer.threadsafe_ack(msg)

        except Exception as e:
            AbstraLogger.error(
                f"[ConsumerController] Error processing message [{msg.delivery_tag}]: {e}"
            )

            AbstraLogger.capture_exception(e)

            self.main_controller.fail_worker_executions(
                app_id=self.app_id,
                worker_id=worker_id,
                reason=f"{e}",
            )

            self.consumer.threadsafe_nack(msg)

            AbstraLogger.warning(
                f"[ConsumerController] Message [{msg.delivery_tag}] has been negatively acknowledged"
            )



from typing import (
    Callable,
    List,
)

from flask import Blueprint, jsonify, make_response, request

from .mcp import create_mcp_tool_handler, register_function


def mcp_bp(functions: List[Callable]) -> Blueprint:
    """Get the editor blueprint with registered MCP tools."""
    editor_bp = Blueprint("mcp", __name__)

    # Store registered tools for MCP discovery - automatically populated
    _registered_tools = {}
    _registered_functions = {}

    # Register tools - metadata is generated automatically
    for func in functions:
        if not callable(func):
            raise TypeError(f"Function {func} is not callable")
        func_name = func.__name__
        if hasattr(func, "_requires_approval"):
            func_name += "__req_approval__"
        register_function(editor_bp, func, _registered_tools, func_name)
        _registered_functions[f"{func_name}"] = func

    # Create dynamic tool handler
    tool_handler = create_mcp_tool_handler(_registered_tools, _registered_functions)

    # Root MCP endpoint for the main protocol communication
    @editor_bp.route("/", methods=["POST", "GET", "OPTIONS"])
    def mcp_root():
        """Main MCP protocol endpoint following JSON-RPC 2.0 and Streamable HTTP"""
        # Handle preflight OPTIONS request
        if request.method == "OPTIONS":
            response = make_response()
            response.headers.add("Access-Control-Allow-Origin", "*")
            response.headers.add(
                "Access-Control-Allow-Headers",
                "Content-Type, Accept, MCP-Protocol-Version, Mcp-Session-Id, Last-Event-ID",
            )
            response.headers.add("Access-Control-Allow-Methods", "POST, GET, OPTIONS")
            response.headers.add("Access-Control-Max-Age", "3600")
            return response

        # Handle GET request for SSE (not implemented but required by spec)
        if request.method == "GET":
            # VS Code might try GET for SSE streams, return 405 as we don't support SSE
            return jsonify({"error": "SSE not supported, use POST"}), 405

        # Handle POST request
        try:
            data = request.get_json(force=True)
            if not data:
                raise ValueError("No JSON data provided")

            # Validate JSON-RPC 2.0 format
            if data.get("jsonrpc") != "2.0":
                raise ValueError("Invalid JSON-RPC version")

            # Handle different MCP method calls
            method = data.get("method")
            request_id = data.get("id")

            if method == "initialize":
                # Check protocol version
                protocol_version = data.get("params", {}).get(
                    "protocolVersion", "2025-06-18"
                )
                supported_versions = ["2025-06-18", "2024-11-05"]

                if protocol_version not in supported_versions:
                    protocol_version = "2025-06-18"  # Default to latest

                response = {
                    "jsonrpc": "2.0",
                    "id": request_id,
                    "result": {
                        "protocolVersion": protocol_version,
                        "capabilities": {"tools": {"listChanged": False}},
                        "serverInfo": {
                            "name": "abstra-mcp-server",
                            "version": "1.0.0",
                            "description": "Abstra workflow management server providing comprehensive tools to inspect, analyze, and debug workflow projects. Supports project structure analysis, stage management, task tracking, and execution monitoring.",
                        },
                        "instructions": "Use the available tools to inspect and analyze Abstra workflow projects. Tools provide comprehensive access to project structure, stages, tasks, executions, and analytics.",
                    },
                }

                # Add proper headers for MCP
                response_obj = jsonify(response)
                response_obj.headers["Content-Type"] = "application/json"
                response_obj.headers["Access-Control-Allow-Origin"] = "*"
                return response_obj

            elif method == "tools/list":
                response = {
                    "jsonrpc": "2.0",
                    "id": request_id,
                    "result": {"tools": list(_registered_tools.values())},
                }
                response_obj = jsonify(response)
                response_obj.headers["Access-Control-Allow-Origin"] = "*"
                return response_obj

            elif method == "notifications/initialized":
                # Handle the initialized notification (no response needed)
                response = make_response("", 204)
                response.headers["Access-Control-Allow-Origin"] = "*"
                return response

            elif method == "tools/call":
                params = data.get("params", {})
                tool_name = params.get("name")
                arguments = params.get("arguments", {})

                try:
                    # Use dynamic tool handler
                    result = tool_handler(tool_name, arguments)

                    response = {"jsonrpc": "2.0", "id": request_id, "result": result}
                    return jsonify(response)
                except ValueError as e:
                    # Unknown tool or invalid tool call
                    error_response = {
                        "jsonrpc": "2.0",
                        "id": request_id,
                        "error": {"code": -32601, "message": str(e)},
                    }
                    return jsonify(error_response), 400
                except TypeError as e:
                    # Invalid parameters (missing required params, wrong types, etc.)
                    error_response = {
                        "jsonrpc": "2.0",
                        "id": request_id,
                        "error": {"code": -32602, "message": str(e)},
                    }
                    return jsonify(error_response), 400

            else:
                error_response = {
                    "jsonrpc": "2.0",
                    "id": request_id,
                    "error": {"code": -32601, "message": f"Unknown method: {method}"},
                }
                return jsonify(error_response), 400

        except Exception as e:
            try:
                data = request.get_json() or {}
            except Exception:
                data = {}
            error_response = {
                "jsonrpc": "2.0",
                "id": data.get("id"),
                "error": {"code": -32603, "message": str(e)},
            }
            print(f"Error processing MCP request: {e}")
            return jsonify(error_response), 500

    # MCP protocol endpoints
    @editor_bp.route("/initialize", methods=["POST"])
    def initialize():
        """MCP initialize endpoint"""
        return jsonify(
            {
                "capabilities": {"tools": {}},
                "serverInfo": {
                    "name": "abstra-mcp-server",
                    "version": "1.0.0",
                    "description": "Abstra workflow management server providing comprehensive tools to inspect, analyze, and debug workflow projects. Supports project structure analysis, stage management, task tracking, and execution monitoring.",
                    "capabilities_description": "Provides read access to all project entities: stages, workflows, tasks, execution logs, and analytics. Enables workflow debugging, task flow analysis, and comprehensive project introspection.",
                },
            }
        )

    @editor_bp.route("/tools/list", methods=["POST"])
    def list_tools():
        """List available MCP tools"""
        return jsonify({"tools": list(_registered_tools.values())})

    @editor_bp.route("/tools/call", methods=["POST"])
    def call_tool():
        """Call an MCP tool"""
        try:
            data = request.get_json()
            tool_name = data.get("name")
            arguments = data.get("arguments", {})

            # Use dynamic tool handler
            result = tool_handler(tool_name, arguments)
            return jsonify(result)

        except ValueError as e:
            return jsonify({"error": str(e)}), 400
        except Exception as e:
            return jsonify({"error": str(e)}), 400

    return editor_bp



from abstra_internals.controllers.sdk.sdk_forms import FormSDKController
from tests.fixtures import BaseTest


class TestFormSDKController(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        mockClient = {}
        mockUsersRepository = {}
        self.controller = FormSDKController(mockClient, mockUsersRepository)  # type: ignore

    def test_reuse(self):
        calls = 0

        def func(a, b, *, c):
            nonlocal calls
            calls += 1
            return a + b + c

        result_a = self.controller.reuse(func, 1, 2, c=3)
        result_b = self.controller.reuse(func, 1, 2, c=3)
        self.assertEqual(result_a, result_b)
        self.assertEqual(calls, 1)

        self.assertEqual(self.controller.reuse(func, 1, 2, c=4), 7)
        self.assertEqual(calls, 2)



import unittest
from datetime import datetime

from abstra_internals.repositories.execution_logs import (
    LocalExecutionLogsRepository,
    LogEntry,
)
from tests.fixtures import clear_dir, init_dir


class TestLocalExecutionLogsRepository(unittest.TestCase):
    def setUp(self) -> None:
        self.root = init_dir()
        self.repository = LocalExecutionLogsRepository()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_starts_empty(self):
        logs = self.repository.get("foo")
        self.assertEqual(len(logs), 0)

    def test_create_stdio(self):
        log_entry = LogEntry(
            execution_id="foo",
            stage_id="bar",
            created_at=datetime.now(),
            event="stdout",
            payload={"text": "hello"},
            sequence=1,
        )
        self.repository.save(log_entry)
        logs = self.repository.get("foo")
        self.assertEqual(len(logs), 1)
        self.assertIsInstance(logs[0], LogEntry)
        self.assertEqual(logs[0].payload["text"], "hello")  # type: ignore // above assertion is enough



import base64
import binascii
import io
import re

import puremagic

BASE_64_PREFIX = "data:image/png;base64,"
PREFFIX_REGEX = re.compile(r"^data:image/(jpeg|png|gif|bmp|webp);base64,")


def is_base_64(unknown_string: str) -> bool:
    if PREFFIX_REGEX.match(unknown_string):
        unknown_string = unknown_string.split(",")[1]
    try:
        base64_bytes = base64.b64decode(unknown_string, validate=True)
        image_type = puremagic.what(None, h=base64_bytes)
        if image_type in ["jpeg", "png", "gif", "bmp", "webp"]:
            return True
        else:
            return False
    except (binascii.Error, ValueError):
        return False


def encode_base_64(input: io.IOBase):
    BASE_64_PREFIX = "data:image/png;base64,"
    input.seek(0)
    file_content = input.read()
    base64_file = base64.b64encode(file_content).decode("utf-8")
    return f"{BASE_64_PREFIX}{base64_file}"



from unittest import TestCase

from circular_imports import cycles_in_path


class TestCircularDeps(TestCase):
    def test_no_cycles(self):
        cycles = cycles_in_path(".", ".venv,build")
        self.assertEqual(len(cycles), 0)



from typing import Callable, List, Optional, Union

from abstra_internals.interface.contract.forms_contract import StepsInfo
from abstra_internals.interface.sdk.forms.deprecated.page import Page
from abstra_internals.interface.sdk.forms.deprecated.page_response import PageResponse
from abstra_internals.interface.sdk.forms.deprecated.steps_response import StepsResponse

Step = Union[Page, Callable]
Steps = List[Step]


def get_page_info(steps: Steps, current_page: Page) -> StepsInfo:
    pages = [step for step in steps if isinstance(step, Page)]
    return StepsInfo(
        current=pages.index(current_page) + 1, total=len(pages), disabled=False
    )


def run_steps(steps: Steps, initial_step=1, context={}) -> StepsResponse:
    assert initial_step >= 1, "Initial step must be greater than or equal to 1"

    executed_steps: Steps = []  # stack
    responses: StepsResponse = StepsResponse()  # stack

    page_steps = [
        (idx, step) for idx, step in enumerate(steps) if isinstance(step, Page)
    ]
    first_step_idx, first_step = page_steps[initial_step - 1]
    steps_to_execute = []
    for idx, step in enumerate(steps):
        if idx < first_step_idx:
            executed_steps.append(step)
            responses.append(context)
        elif idx == first_step_idx:
            first_step = step
        elif idx > first_step_idx:
            steps_to_execute.append(step)

    steps_to_execute.reverse()

    assert first_step is not None
    if not isinstance(first_step, Page):
        raise ValueError("First step needs to be a Page")
    response = run_page(steps, first_step, executed_steps, responses, context)

    while steps_to_execute or response.action == "i18n_back_action":
        if steps_to_execute and response.action != "i18n_back_action":
            next_page = execute_functions(steps_to_execute, executed_steps, responses)
            if not next_page:
                break
            response = run_page(steps, next_page, executed_steps, responses, context)

        if response.action == "i18n_back_action":
            response = go_back(steps, steps_to_execute, executed_steps, responses)

    return responses


def go_back(
    steps: Steps,
    steps_to_execute: Steps,
    executed_steps: Steps,
    responses: StepsResponse,
):
    page = executed_steps.pop()
    responses.pop()
    steps_to_execute.append(page)

    next_page = undo_functions(steps_to_execute, executed_steps, responses)
    response = run_back_page(steps, next_page, executed_steps, responses)
    return response


def execute_functions(
    steps_to_execute: Steps, executed_steps: Steps, responses: StepsResponse
) -> Optional[Page]:
    while steps_to_execute:
        step = steps_to_execute.pop()
        if not isinstance(step, Page):
            responses.append(step(responses))
            executed_steps.append(step)
        else:
            return step
    return None


def undo_functions(
    steps_to_execute: Steps, executed_steps: Steps, responses: StepsResponse
) -> Page:
    while executed_steps:
        step = executed_steps.pop()
        if not isinstance(step, Page):
            steps_to_execute.append(step)
            responses.pop()
        else:
            return step

    raise ValueError("No steps to undo")


def run_page(
    steps: Steps,
    next_page: Page,
    executed_steps: Steps,
    responses: StepsResponse,
    context: dict,
) -> PageResponse:
    steps_info = get_page_info(steps, next_page)
    response = (
        next_page.run(steps_info=steps_info, context=responses.acc)
        if steps_info["current"] == 1
        else next_page.run(
            actions=["i18n_back_action", "i18n_next_action"],
            steps_info=steps_info,
            context=responses.acc,
        )
    )
    responses.append(response)
    executed_steps.append(next_page)
    return response


def run_back_page(
    steps: Steps,
    next_page: Page,
    executed_steps: Steps,
    responses: StepsResponse,
):
    steps_info = get_page_info(steps, next_page)
    responses.pop()
    response = (
        next_page.run(context=responses.acc, steps_info=steps_info)
        if steps_info["current"] == 1
        else next_page.run(
            actions=["i18n_back_action", "i18n_next_action"],
            context=responses.acc,
            steps_info=steps_info,
        )
    )
    responses.append(response)
    executed_steps.append(next_page)
    return response



from inspect import stack
from pathlib import Path
from unittest import TestCase, mock

from abstra_internals.debug import make_exception_debug_data, make_frame_debug_data
from abstra_internals.settings import Settings
from tests.resources.make_debug.module import func


class TestDebug(TestCase):
    def setUp(self) -> None:
        Settings.set_root_path(str(Path(__file__).parent))

    def test_make_debug_data_from_exception(self):
        try:
            raise Exception("test")
        except Exception as exc:
            debug_data = make_exception_debug_data(exc)

        import os

        expected_filename = os.path.normcase(os.path.normpath(__file__))

        self.assertEqual(
            debug_data,
            dict(
                debug=dict(
                    stack=[
                        dict(
                            filename=expected_filename,
                            lineno=16,
                            name="test_make_debug_data_from_exception",
                            locals=mock.ANY,
                        )
                    ]
                )
            ),
        )

    def test_make_debug_data_from_frames(self):
        debug_data = make_frame_debug_data(stack())

        import os

        expected_filename = os.path.normcase(os.path.normpath(__file__))

        self.assertEqual(
            debug_data,
            dict(
                debug=dict(
                    stack=[
                        dict(
                            filename=expected_filename,
                            lineno=41,
                            name="test_make_debug_data_from_frames",
                            locals=mock.ANY,
                        )
                    ]
                )
            ),
        )

    def test_make_debug_data_nested(self):
        def outer_func():
            raise Exception("test")

        try:
            func(outer_func)
            debug_data = None
        except Exception as exc:
            debug_data = make_exception_debug_data(exc)

        import os

        expected_filename = os.path.normcase(os.path.normpath(__file__))
        expected_module_filename = os.path.normcase(
            os.path.normpath(
                str((Path(__file__) / "../resources/make_debug/module.py").resolve())
            )
        )

        self.assertEqual(
            debug_data,
            dict(
                debug=dict(
                    stack=[
                        dict(
                            filename=expected_filename,
                            lineno=68,
                            name="test_make_debug_data_nested",
                            locals=mock.ANY,
                        ),
                        dict(
                            filename=expected_module_filename,
                            lineno=2,
                            name="func",
                            locals=mock.ANY,
                        ),
                        dict(
                            filename=expected_filename,
                            lineno=65,
                            name="outer_func",
                            locals=mock.ANY,
                        ),
                    ]
                )
            ),
        )

    def test_make_debug_data_nested_other_dir(self):
        Settings.set_root_path("./resources")

        def outer_func():
            raise Exception("test")

        try:
            func(outer_func)
            debug_data = None
        except Exception as exc:
            debug_data = make_exception_debug_data(exc)

        import os

        expected_module_filename = os.path.normcase(
            os.path.normpath(
                str((Path(__file__) / "../resources/make_debug/module.py").resolve())
            )
        )

        self.assertEqual(
            debug_data,
            dict(
                debug=dict(
                    stack=[
                        dict(
                            filename=expected_module_filename,
                            lineno=2,
                            name="func",
                            locals=mock.ANY,
                        ),
                    ]
                )
            ),
        )



import shutil
import unittest
from pathlib import Path

from abstra_internals.repositories.execution import ExecutionFilter
from abstra_internals.repositories.factory import build_editor_repositories
from abstra_internals.repositories.multiprocessing import SpawnContextReposity
from abstra_internals.settings import Settings


def worker_function(test_dir: str):
    """Function executed in the child process."""
    Settings.set_root_path(test_dir)

    # This line used to cause the SemLock error before the fix
    repositories = build_editor_repositories()

    # Test basic SqlStorage operation
    filter = ExecutionFilter()
    repositories.execution.list(filter)

    return True


class TestSemLockSimple(unittest.TestCase):
    """Simple and direct test for the SemLock fix.

    Reproduces exactly the problem:
    1. Creates SqlStorage in PARENT process with threading.RLock
    2. Tries to spawn CHILD process
    3. CHILD process uses SqlStorage

    If there was a SemLock error, it would fail at step 2 (pickle of spawn).
    """

    def setUp(self):
        self.test_dir = Path(__file__).parent / ".abstra_test_simple"
        self.test_dir.mkdir(exist_ok=True)
        Settings.set_root_path(str(self.test_dir))

    def tearDown(self):
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)

    def test_spawn_process_with_repositories(self):
        """Verifies that the SemLock fix is working"""
        # Create repositories in PARENT process
        repositories = build_editor_repositories()
        self.assertIsNotNone(repositories)

        # Try to spawn CHILD process
        mp_context = SpawnContextReposity().get_context()

        # This spawn would cause SemLock error if RLock was multiprocessing.RLock
        process = mp_context.Process(target=worker_function, args=(str(self.test_dir),))

        process.start()
        process.join(timeout=10)

        # Verify that the process completed successfully
        self.assertEqual(
            process.exitcode,
            0,
            "Child process should have finished with exitcode 0 (no SemLock error)",
        )



from pathlib import Path

from abstra_internals.controllers.execution.execution import ExecutionController
from abstra_internals.controllers.execution.execution_client_hook import HookClient
from abstra_internals.entities.execution_context import HookContext, Request, Response
from abstra_internals.repositories.project.project import HookStage
from tests.fixtures import BaseTest


class ExecutionControllerTest(BaseTest):
    def setUp(self) -> None:
        super().setUp()

        self.request = Request(
            body={"a": 1}.__str__(),
            headers={"auth": "some_secret_token"},
            query_params={"c": "3"},
            method="GET",
        )

        self.context = HookContext(
            request=Request(
                body={"a": 1}.__str__(),
                headers={"auth": "some_secret_token"},
                query_params={"c": "3"},
                method="GET",
            ),
            response=Response(headers={}, status=200, body=""),
        )

        self.project = self.repositories.project.load(include_disabled_stages=False)
        self.stage = HookStage.create(
            title="mock_stage",
            file="mock_file.py",
            id="mock_hook_id",
            workflow_position=(0, 1),
        )
        Path(self.stage.file).write_text("print('Hello, World!')", encoding="utf-8")
        self.project.add_stage(self.stage)
        self.repositories.project.save(self.project)

        self.hook_client = HookClient(self.context)

    def test_run_initial_returns_dto(self):
        ExecutionController(
            repositories=self.repositories,
            stage=self.stage,
            context=self.context,
            client=self.hook_client,
        ).run()

        if not self.context.response:
            self.fail("Response was not set")

        self.assertEqual(self.context.response.status, 200)



import sys
import threading
from typing import Callable, List, Literal, Optional, Union

import flask_sock

from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.sdk.sdk_context import (
    SDKContextStore,
)
from abstra_internals.entities.execution import Execution
from abstra_internals.env_masker import GLOBAL_MASKER
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.interface.sdk.user_exceptions import ExecutionNotFound
from abstra_internals.logger import AbstraLogger


class BroadcastController:
    listeners: List[flask_sock.Server] = []
    _lock = threading.Lock()

    @classmethod
    def register(cls, listener: flask_sock.Server):
        with cls._lock:
            cls.listeners.append(listener)

    @classmethod
    def unregister(cls, listener: flask_sock.Server):
        with cls._lock:
            if listener in cls.listeners:
                cls.listeners.remove(listener)

    @classmethod
    def broadcast(cls, *, msg: str):
        for listener in cls.listeners:
            try:
                listener.send(msg)
            except Exception:
                BroadcastController.unregister(listener)

    def __init__(
        self,
        *,
        main_controller: MainController,
        sys_stdout_write,
        sys_stderr_write,
    ):
        self.execution_logs_repository = main_controller.execution_logs_repository
        self.execution_repository = main_controller.execution_repository
        self.sys_stdout_write = sys_stdout_write
        self.sys_stderr_write = sys_stderr_write

    def patched_stderr_write(self, raw: Union[str, bytearray]) -> int:
        return self._handle_stdio("stderr", self.sys_stderr_write, raw)

    def patched_stdout_write(self, raw: Union[str, bytearray]) -> int:
        return self._handle_stdio("stdout", self.sys_stdout_write, raw)

    def _handle_stdio(
        self,
        std_type: Literal["stdout", "stderr"],
        sys_write: Callable,
        raw: Union[str, bytearray],
    ):
        text = raw.decode("utf-8") if not isinstance(raw, str) else raw

        try:
            execution = self.get_current_execution()

            text = self.mask(text)
            text = self.tag(text, execution)

            self.send_stdio(execution, std_type, text)
        except Exception as e:
            AbstraLogger.capture_exception(e)
        finally:
            sys_write(text)
            sys.stdout.flush()
            return len(text)

    def get_current_execution(self) -> Optional[Execution]:
        try:
            return SDKContextStore.get_execution()
        except ExecutionNotFound:
            return None

    def send_stdio(
        self,
        execution: Optional[Execution],
        std_type: Literal["stderr", "stdout"],
        text: str,
    ):
        if not execution:
            return

        self.execution_logs_repository.insert_stdio(
            execution.id, execution.stage_id, std_type, text
        )

    def mask(self, raw: str) -> str:
        if IS_PRODUCTION:
            return GLOBAL_MASKER.mask(raw)
        return raw

    def tag(self, raw: str, execution: Optional[Execution]) -> str:
        if not execution:
            return raw

        if not raw.strip():
            return raw

        short_id = execution.id.split(sep="-")[0]
        prefix = f"[RUN {short_id}] "

        lines = raw.splitlines(keepends=True)
        if len(lines) <= 1:
            return f"{prefix}{raw}"

        tagged_lines = []
        for line in lines:
            tagged_lines.append(f"{prefix}{line}")

        return "".join(tagged_lines)



import asyncio
import json
import socket
import uuid
from threading import Thread
from typing import Any, Optional

import requests
import websockets
from colorama import Fore
from pydantic import BaseModel

from abstra_internals.contracts_generated import (
    CloudApiCliApiKeyInfoResponse,
    CloudApiCliBuildCreateResponse,
)
from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import (
    CLOUD_API_CLI_URL,
    CLOUD_API_ENDPOINT,
    HOST,
    REQUEST_TIMEOUT,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.settings import Settings


def create_build(headers: dict) -> CloudApiCliBuildCreateResponse:
    url = f"{CLOUD_API_CLI_URL}/builds"
    r = requests.post(url=url, headers=headers, timeout=REQUEST_TIMEOUT)
    if not r.ok:
        raise Exception(f"Failed to create build: {r.text}")

    data = r.json()
    return CloudApiCliBuildCreateResponse.from_dict(data)


def update_build(build_id: str, headers: dict, error: Optional[str] = None):
    url = f"{CLOUD_API_CLI_URL}/builds/{build_id}"
    requests.patch(
        url=url, headers=headers, json=dict(error=error), timeout=REQUEST_TIMEOUT
    ).raise_for_status()


def get_api_key_info(headers: dict) -> dict:
    url = f"{CLOUD_API_CLI_URL}/api-keys/info"
    try:
        response = requests.get(url, headers=headers, timeout=REQUEST_TIMEOUT)
    except Exception as e:
        AbstraLogger.capture_exception(e)
        return {"logged": False, "reason": "CONNECTION_ERROR"}

    if response.ok:
        response_data = CloudApiCliApiKeyInfoResponse.from_dict(response.json())
        return {"logged": True, "info": response_data.to_dict()}
    else:
        return {"logged": False, "reason": "INVALID_API_TOKEN"}


def get_project_info(headers: dict, project_id: Optional[str] = None):
    if project_id:
        url = f"{CLOUD_API_CLI_URL}/project/{project_id}"
    else:
        url = f"{CLOUD_API_CLI_URL}/project"
    r = requests.get(url, headers=headers, timeout=REQUEST_TIMEOUT)
    r.raise_for_status()
    return r.json()


def get_feature_flags(headers: dict):
    url = f"{CLOUD_API_CLI_URL}/project/feature-flags"
    r = requests.get(url, headers=headers, timeout=REQUEST_TIMEOUT)
    r.raise_for_status()
    return r.json()


class TunnelRequest(BaseModel):
    method: str
    path: str
    headers: dict
    body: Optional[dict]
    query: dict
    sessionPath: str
    requestId: str


class TunnelResponse(BaseModel):
    status: int
    headers: dict
    text: str
    sessionPath: str
    requestId: str


class SessionPathMessage(BaseModel):
    sessionPath: str


def connect_tunnel(verbose: bool = False):
    url = f"{CLOUD_API_ENDPOINT}/tunnel/connect?format=dict".replace(
        "https://", "wss://"
    ).replace("http://", "ws://")

    async def loop():
        await asyncio.sleep(1)
        while True:
            headers = resolve_headers()
            if headers is None:
                await asyncio.sleep(5)
                continue
            try:
                async with websockets.connect(
                    uri=url,
                    additional_headers=headers,
                    ping_interval=20,
                    ping_timeout=10,
                    close_timeout=10,
                ) as ws:
                    if verbose:
                        print("WebSocket connection established")
                    while True:
                        try:
                            message = await asyncio.wait_for(ws.recv(), timeout=30.0)
                            assert isinstance(message, str)
                            message_dict = json.loads(message)

                            if "method" in message_dict:
                                request = TunnelRequest.model_validate_json(message)

                                cookies = {}
                                editor_auth_token = get_editor_auth_token_from_file()
                                if editor_auth_token:
                                    cookies["editor_auth"] = editor_auth_token

                                kwargs: Any = dict(
                                    headers=request.headers,
                                    params=request.query,
                                    cookies=cookies,
                                )

                                kwargs["json"] = request.body
                                if (
                                    not request.path.startswith("/_hooks/")
                                    and not request.path.startswith("/_healthcheck")
                                    and not request.path.startswith("/_tasks")
                                    and not request.path.startswith("/_editor/api/mcp")
                                ):
                                    response = TunnelResponse(
                                        status=403,
                                        headers={},
                                        text=f"Forbidden path: {request.path}",
                                        sessionPath=request.sessionPath,
                                        requestId=request.requestId,
                                    )
                                    await ws.send(response.model_dump_json())
                                else:
                                    response = requests.request(
                                        request.method,
                                        f"http://{HOST}:{Settings.server_port}{request.path}",
                                        timeout=REQUEST_TIMEOUT,
                                        **kwargs,
                                    )
                                    response_json = TunnelResponse(
                                        status=response.status_code,
                                        headers=dict(response.headers),
                                        text=response.text,
                                        sessionPath=request.sessionPath,
                                        requestId=request.requestId,
                                    ).model_dump_json()

                                    await ws.send(response_json)

                            else:
                                global session
                                session = SessionPathMessage.model_validate_json(
                                    message
                                )
                                set_session_path(session.sessionPath)
                                public_url = f"{CLOUD_API_ENDPOINT}/tunnel/forward/{session.sessionPath}"
                                Settings.set_public_url(public_url)

                                if not hasattr(loop, "_printed"):
                                    setattr(loop, "_printed", True)
                                    print(
                                        f"You can test your hooks locally by sending requests to: {Fore.GREEN}{public_url}/_hooks/:hook-path{Fore.RESET}"
                                    )
                        except asyncio.TimeoutError:
                            try:
                                await ws.ping()
                                continue
                            except Exception:
                                if verbose:
                                    print("WebSocket ping failed, connection lost")
                                break
                        except websockets.exceptions.ConnectionClosedOK:
                            if verbose:
                                print("WebSocket connection closed OK")
                            await asyncio.sleep(5)
                            break
                        except websockets.exceptions.ConnectionClosedError as e:
                            if verbose:
                                print(f"WebSocket connection closed with error: {e}")
                            await asyncio.sleep(5)
                            break
            except socket.gaierror:
                if verbose:
                    print(
                        "Network error: Unable to resolve host. Retrying in 5 seconds..."
                    )
                await asyncio.sleep(5)
                continue
            except ConnectionRefusedError:
                if verbose:
                    print("Network error: Connection refused. Retrying in 5 seconds...")
                await asyncio.sleep(5)
                continue
            except Exception as e:
                AbstraLogger.capture_exception(e)
                if verbose:
                    print(f"Unexpected error: {e}. Retrying in 5 seconds...")
                await asyncio.sleep(5)
                continue

    def run_async_loop():
        asyncio.run(loop())

    Thread(target=run_async_loop, daemon=True, name="TunnelLoop").start()


def get_session_path() -> str:
    path = Settings.root_path / ".abstra" / "session_path"
    assert path.exists(), f"Session path file does not exist: {path}"
    return path.read_text(encoding="utf-8").strip()


def set_session_path(session_path: str):
    path = Settings.root_path / ".abstra" / "session_path"
    path.write_text(session_path, encoding="utf-8")
    refresh_tunnel_secret_key()


def refresh_tunnel_secret_key():
    secret_key = uuid.uuid4().hex
    path = Settings.root_path / ".abstra" / "secret_key"
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(secret_key, encoding="utf-8")


def get_tunnel_secret_key() -> str:
    path = Settings.root_path / ".abstra" / "secret_key"
    if not path.exists():
        refresh_tunnel_secret_key()
    return path.read_text(encoding="utf-8").strip()


def get_editor_auth_token_from_file() -> str:
    path = Settings.root_path / ".abstra" / "editor_auth_token"
    if not path.exists():
        return ""
    return path.read_text(encoding="utf-8").strip()


def save_editor_auth_token_to_file(token: str):
    path = Settings.root_path / ".abstra" / "editor_auth_token"
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(token, encoding="utf-8")



import userpath
from pip._internal.locations import get_bin_prefix, get_bin_user  # type: ignore


def main():
    bin_prefix = get_bin_prefix()
    bin_user = get_bin_user()

    if not userpath.in_current_path(bin_user):
        print(f"Adding {bin_user} to PATH")
        userpath.append(bin_user)

    if not userpath.in_current_path(bin_prefix):
        print(f"Adding {bin_prefix} to PATH")
        userpath.append(bin_prefix)

    if userpath.need_shell_restart(bin_user) or userpath.need_shell_restart(bin_prefix):
        print("Please restart your shell to apply changes")
    else:
        print("PATH is already up-to-date")


if __name__ == "__main__":
    main()

__all__ = []



from abstra_internals.entities.forms.widgets.widget_base import OutputWidget


class MarkdownOutput(OutputWidget):
    """Markdown output widget for displaying formatted text.

    Attributes:
        text (str): The markdown content to display.
    """

    type = "markdown-output"

    def __init__(
        self,
        text: str,
        *,
        full_width: bool = False,
    ):
        """Initialize a MarkdownOutput widget.

        Args:
            text (str): The markdown content to display.
            full_width (bool): Whether the content should take up the full width of its container.
        """
        self.text = text
        self.full_width = full_width

    def _render(self):
        return {"type": self.type, "text": self.text, "fullWidth": self.full_width}



import os
import ssl
import subprocess
import sys
import threading

import certifi
from dotenv import load_dotenv
from werkzeug.serving import make_server

from abstra_internals.cloud_api import connect_tunnel
from abstra_internals.controllers.codebase_events import CodebaseEventController
from abstra_internals.controllers.execution.consumer import ConsumerController
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import HOST
from abstra_internals.interface.cli.messages import serve_message
from abstra_internals.logger import AbstraLogger
from abstra_internals.logs_watcher import LogsWatcher, on_logs_update
from abstra_internals.repositories.consumer import EditorConsumer
from abstra_internals.repositories.factory import build_editor_repositories
from abstra_internals.repositories.producer import LocalProducerRepository
from abstra_internals.server.apps import get_local_app
from abstra_internals.services.file_watcher import FileWatcher
from abstra_internals.settings import Settings
from abstra_internals.stdio_patcher import StdioPatcher
from abstra_internals.tasks_watcher import TasksWatcher, on_tasks_update
from abstra_internals.utils.browser import background_open_editor
from abstra_internals.version import check_latest_version


def start_consumer(controller: MainController):
    if not isinstance(controller.producer_repository, LocalProducerRepository):
        raise ValueError("Invalid producer repository")

    consumer = EditorConsumer(controller.producer_repository.queue)

    th = threading.Thread(
        daemon=True,
        name="ExecutionConsumer",
        target=ConsumerController(controller, consumer).start_loop,
    )

    th.start()

    return consumer, th


def ensure_certificates():
    try:
        cafile = certifi.where()
        if not os.path.isfile(cafile):
            raise FileNotFoundError(f"Certifi CA file not found at {cafile}")
        ssl.create_default_context(cafile=cafile)
    except Exception:
        print(
            "SSL certificate validation failed. Attempting to restore certificates..."
        )
        try:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--force-reinstall",
                    "certifi",
                ],
                check=True,
            )
            import certifi as certifi_reinstalled

            cafile_re = certifi_reinstalled.where()
            if os.path.isfile(cafile_re):
                print(f"Certificates restored at: {cafile_re}")
            else:
                print(
                    f"Certifi CA file still missing at: {cafile_re}. Please check your Python environment."
                )
        except Exception as update_e:
            print(f"Failed to restore certificates: {update_e}")


def editor(headless: bool, verbose: bool = False):
    ensure_certificates()

    load_dotenv(Settings.root_path / ".env")

    serve_message()
    check_latest_version()
    AbstraLogger.init("local")

    repositories = build_editor_repositories()
    main_controller = MainController(repositories)
    main_controller.reset_repositories()
    StdioPatcher.apply(main_controller)

    codebase_event_controller = CodebaseEventController(repositories)
    watcher = FileWatcher(
        [
            codebase_event_controller.reload_env,
            codebase_event_controller.reload_modules,
            codebase_event_controller.lint_files,
            codebase_event_controller.broadcast_changes,
        ]
    )
    watcher.start()

    logs_watcher = LogsWatcher([on_logs_update])
    logs_watcher.start()

    tasks_watcher = TasksWatcher([on_tasks_update])
    tasks_watcher.start()

    start_consumer(main_controller)

    app = get_local_app(main_controller)
    server = make_server(host=HOST, port=Settings.server_port, threaded=True, app=app)

    if not headless:
        background_open_editor()

    connect_tunnel(verbose=verbose)

    server.serve_forever()



import abc
import typing
from dataclasses import dataclass
from typing import Optional

from abstra_internals.cloud_api.http_client import AbstraHTTPResponse, HTTPClient
from abstra_internals.environment import REQUEST_TIMEOUT


@dataclass
class ColumnDTO:
    id: str
    name: str
    type: str
    table_id: str
    unique: bool
    default: str
    nullable: bool
    primary_key: bool
    protected: bool
    foreign_key: Optional[str]

    @staticmethod
    def from_dict(data: dict) -> "ColumnDTO":
        return ColumnDTO(
            id=data["id"],
            name=data["name"],
            type=data["type"],
            table_id=data["tableId"],
            unique=data["unique"],
            default=data["default"],
            nullable=data["nullable"],
            primary_key=data["primaryKey"],
            protected=data["protected"],
            foreign_key=data["foreignKey"],
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class TableDTO:
    id: str
    name: str
    project_id: str

    @staticmethod
    def from_dict(data: dict) -> "TableDTO":
        return TableDTO(
            id=data["id"],
            name=data["name"],
            project_id=data["projectId"],
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


class TablesRepository(abc.ABC):
    client: "HTTPClient"

    def __init__(self, client: "HTTPClient") -> None:
        self.client = client

    def execute(self, query: str, params: typing.List) -> AbstraHTTPResponse:
        raise NotImplementedError()

    def create_table(self, id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def create_column(self, table_id: str, name: str, type: str) -> ColumnDTO:
        raise NotImplementedError()

    def insert_row(self, table_id: str, values: dict):
        raise NotImplementedError()

    def update_table(self, table_id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def update_column(self, column_id: str, table_id: str, changes: dict) -> ColumnDTO:
        raise NotImplementedError()


class ProductionTablesRepository(TablesRepository):
    def execute(self, query: str, params: typing.List) -> AbstraHTTPResponse:
        body = {"query": query, "params": params}
        return self.client.post(
            "/tables/execute",
            json=body,
        )

    def create_table(self, id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def create_column(self, table_id: str, name: str, type: str) -> ColumnDTO:
        raise NotImplementedError()

    def insert_row(self, table_id: str, values: dict):
        raise NotImplementedError()

    def update_table(self, table_id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def update_column(self, column_id: str, table_id: str, changes: dict) -> ColumnDTO:
        raise NotImplementedError()


class LocalTablesRepository(TablesRepository):
    def execute(self, query: str, params: typing.List) -> AbstraHTTPResponse:
        body = {"query": query, "params": params}

        return self.client.post("/tables/execute", json=body, timeout=REQUEST_TIMEOUT)

    def create_table(self, id: str, name: str) -> TableDTO:
        r = self.client.post(
            "/table",
            json={
                "id": id,
                "name": name,
            },
        )
        table = r.json()["response"]["table"]
        return TableDTO.from_dict(table)

    def create_column(self, table_id: str, name: str, type: str) -> ColumnDTO:
        body = {
            "tableId": table_id,
            "name": name,
            "type": type,
        }
        r = self.client.post("/tables/column", json=body)
        column = r.json()["response"]
        return ColumnDTO.from_dict(column)

    def insert_row(self, table_id: str, values: dict):
        body = {
            "tableId": table_id,
            "row": values,
        }
        self.client.post("/tables/row", json=body)

    def update_table(self, table_id: str, name: str) -> TableDTO:
        body = {
            "name": name,
        }
        r = self.client.patch(f"/tables/table/{table_id}", json=body)
        table = r.json()["response"]
        return TableDTO.from_dict(table)

    def update_column(self, column_id: str, table_id: str, changes: dict) -> ColumnDTO:
        body = {
            "tableId": table_id,
            "changes": changes,
        }
        r = self.client.patch(f"/tables/column/{column_id}", json=body)
        column = r.json()["response"]
        return ColumnDTO.from_dict(column)



import shutil
import subprocess
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from abstra_internals.environment import REMOTE_NAME

from .types import GitCommit, GitFileChange, GitRepositoryInterface, GitStatus

TEMP_ABSTRA_EMAIL = "abstra@abstra.app"
TEMP_ABSTRA_NAME = "Abstra"


class NativeGitRepository(GitRepositoryInterface):
    """Repository for Git operations using system git command"""

    def __init__(self, working_directory: Path):
        super().__init__(working_directory)
        self._git_available = None

    def _detect_and_cleanup_stale_locks(self) -> bool:
        """Detect and clean up stale git lock files."""
        if not self.is_git_repository():
            return False

        cleaned = False
        lock_patterns = [
            self.working_directory / ".git" / "index.lock",
            self.working_directory / ".git" / "HEAD.lock",
            self.working_directory / ".git" / "config.lock",
        ]

        refs_dir = self.working_directory / ".git" / "refs"
        if refs_dir.exists():
            lock_patterns.extend(refs_dir.rglob("*.lock"))

        for lock_file in lock_patterns:
            if isinstance(lock_file, Path) and lock_file.exists():
                try:
                    lock_file.unlink()
                    cleaned = True
                    print(
                        f"Cleaned up stale lock file: {lock_file.relative_to(self.working_directory)}"
                    )
                except Exception as e:
                    print(f"Failed to clean lock file {lock_file}: {e}")

        return cleaned

    def _handle_lock_file_error(self, stderr: str) -> Tuple[bool, str]:
        """Detect if error is due to lock files and attempt to clean them up."""
        lock_error_indicators = [
            "index.lock",
            ".lock': File exists",
            "Unable to create",
            "another git process seems to be running",
            "couldn't set 'refs/heads",
        ]

        is_lock_error = any(
            indicator.lower() in stderr.lower() for indicator in lock_error_indicators
        )

        if is_lock_error:
            print("Detected stale lock file error. Attempting cleanup...")
            cleaned = self._detect_and_cleanup_stale_locks()

            if cleaned:
                return (
                    True,
                    "Lock files were cleaned up. Please try your operation again.",
                )
            else:
                return False, (
                    "Git lock file detected but couldn't be cleaned automatically. "
                    "This usually means another git operation is in progress. "
                    "If you're certain no other git operation is running, you may need to manually "
                    "delete .lock files in the .git directory."
                )

        return False, stderr

    def _run_git_command(
        self,
        command: List[str],
        cwd: Optional[Path] = None,
        input: Optional[str] = None,
        auto_cleanup_locks: bool = True,
        timeout: Optional[int] = None,
    ) -> Tuple[bool, str, str]:
        """Run a git command and return success, stdout, stderr"""
        process = None
        try:
            if cwd is None:
                cwd = self.working_directory

            if timeout is None:
                network_operations = ["push", "pull", "fetch", "clone"]
                if any(op in command for op in network_operations):
                    timeout = 120
                else:
                    timeout = 30

            process = subprocess.Popen(
                ["git"] + command,
                cwd=cwd,
                stdin=subprocess.PIPE if input else None,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )

            stdout, stderr = process.communicate(input=input, timeout=timeout)
            success = process.returncode == 0

            if not success and auto_cleanup_locks:
                should_retry, _ = self._handle_lock_file_error(stderr)
                if should_retry:
                    return self._run_git_command(
                        command, cwd, input, auto_cleanup_locks=False, timeout=timeout
                    )

            return success, stdout.strip(), stderr.strip()

        except subprocess.TimeoutExpired:
            if process:
                try:
                    process.terminate()
                    try:
                        process.wait(timeout=2)
                    except subprocess.TimeoutExpired:
                        process.kill()
                        process.wait()
                except Exception:
                    pass

            timeout_msg = f"Git command timed out after {timeout} seconds"
            return False, "", timeout_msg

        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            return False, "", str(e)

        except Exception as e:
            if process and process.poll() is None:
                try:
                    process.kill()
                    process.wait()
                except Exception:
                    pass
            return False, "", str(e)

    def is_git_available(self) -> bool:
        """Check if git is installed and available"""
        if self._git_available is None:
            self._git_available = shutil.which("git") is not None
        return self._git_available

    def find_git_root(self, start_path: Optional[Path] = None) -> Optional[Path]:
        """Find the root directory of the git repository"""
        if not self.is_git_available():
            return None

        if start_path is None:
            start_path = self.working_directory

        try:
            success, stdout, _ = self._run_git_command(
                ["rev-parse", "--show-toplevel"], cwd=start_path
            )
            if success and stdout:
                return Path(stdout.strip()).resolve()
        except Exception:
            return None

    def is_git_repository(self) -> bool:
        """Check if current directory is a git repository"""
        if not self.is_git_available():
            return False

        success, _, _ = self._run_git_command(["rev-parse", "--git-dir"])
        return success

    def get_current_branch(self) -> Optional[str]:
        """Get the current branch name"""
        if not self.is_git_repository():
            return None

        success, stdout, _ = self._run_git_command(["branch", "--show-current"])
        if success and stdout:
            return stdout

        success, stdout, _ = self._run_git_command(["symbolic-ref", "--short", "HEAD"])
        if success and stdout:
            return stdout

        return None

    def get_all_branches(self) -> List[str]:
        """Get all local branches"""
        if not self.is_git_repository():
            return []

        success, stdout, _ = self._run_git_command(
            ["branch", "--format=%(refname:short)"]
        )
        if success and stdout:
            return [branch.strip() for branch in stdout.split("\n") if branch.strip()]

        # Fallback for repositories without commits: check if HEAD points to a branch
        # This handles the case where a repository was initialized but no commits were made yet
        head_success, head_output, _ = self._run_git_command(
            ["symbolic-ref", "--short", "HEAD"]
        )
        if head_success and head_output.strip():
            return [head_output.strip()]

        return []

    def get_last_commit(self) -> Optional[GitCommit]:
        """Get information about the last commit"""
        commits = self.get_commit_history(limit=1)
        return commits[0] if commits else None

    def get_commit_history(
        self, limit: int = 10, offset: int = 0, branch: Optional[str] = None
    ) -> List[GitCommit]:
        """Get commit history with pagination"""
        if not self.is_git_repository():
            return []

        if branch:
            git_args = ["log", branch]
        else:
            git_args = ["log", "--all"]

        git_args.extend(
            [
                f"--skip={offset}",
                f"--max-count={limit}",
                "--format=%H|%s|%an|%ad",
                "--date=iso",
            ]
        )

        success, stdout, _ = self._run_git_command(git_args)

        if success and stdout:
            commits = []
            for line in stdout.strip().split("\n"):
                if line.strip():
                    parts = line.split("|", 3)
                    if len(parts) == 4:
                        commits.append(
                            GitCommit(
                                hash=parts[0],
                                message=parts[1],
                                author=parts[2],
                                date=parts[3],
                            )
                        )
            return commits
        return []

    def get_changed_files(self) -> List[str]:
        """Get list of changed files"""
        if not self.is_git_repository():
            return []

        success, stdout, _ = self._run_git_command(["status", "--porcelain"])
        if success and stdout:
            files = []
            for line in stdout.split("\n"):
                if line.strip():
                    filename = line[2:].lstrip() if len(line) > 2 else ""
                    if filename:
                        files.append(filename)
            return files
        return []

    def get_changed_files_with_status(self) -> List[GitFileChange]:
        """Get list of changed files with their git status"""
        if not self.is_git_repository():
            return []

        success, stdout, _ = self._run_git_command(["status", "--porcelain"])
        if success and stdout:
            files = []
            for line in stdout.split("\n"):
                if line.strip():
                    if len(line) < 3:
                        continue

                    status_code = line[:2]
                    filename = line[2:].lstrip() if len(line) > 2 else ""

                    if not filename:
                        continue

                    if status_code[0] == "A" or status_code[1] == "A":
                        status = "added"
                    elif status_code[0] == "D" or status_code[1] == "D":
                        status = "deleted"
                    elif status_code[0] == "M" or status_code[1] == "M":
                        status = "modified"
                    elif status_code[0] == "R" or status_code[1] == "R":
                        status = "renamed"
                    elif status_code[0] == "?" and status_code[1] == "?":
                        status = "untracked"
                    else:
                        status = "modified"

                    files.append(
                        GitFileChange(
                            path=filename, status=status, status_code=status_code
                        )
                    )
            return files
        return []

    def has_uncommitted_changes(self) -> bool:
        """Check if there are uncommitted changes"""
        return len(self.get_changed_files()) > 0

    def init_repository(self) -> bool:
        """Initialize a new git repository in the working directory"""
        if not self.is_git_available():
            return False

        # Initialize repository with main as default branch
        success, _, _ = self._run_git_command(["init", "-b", "main"])
        if not success:
            # Fallback for older git versions that don't support -b flag
            success, _, _ = self._run_git_command(["init"])
            if not success:
                return False
            # Set main as default branch
            self._run_git_command(["symbolic-ref", "HEAD", "refs/heads/main"])

        self.configure_git_user(TEMP_ABSTRA_EMAIL, TEMP_ABSTRA_NAME)

        add_success, _, _ = self._run_git_command(["add", "."])
        if add_success:
            status_success, status_output, _ = self._run_git_command(
                ["status", "--porcelain", "--cached"]
            )
            if status_success and status_output.strip():
                commit_success, _, _ = self._run_git_command(
                    ["commit", "-m", "First commit"]
                )
                if commit_success:
                    return True

        # If no files to commit or commit failed, create an empty initial commit
        # This ensures we have a branch with at least one commit
        empty_commit_success, _, _ = self._run_git_command(
            ["commit", "--allow-empty", "-m", "First commit"]
        )

        return empty_commit_success

    def configure_git_user(self, fallback_email: str, fallback_name: str):
        """Ensure git user is configured for commits (required in CI environments)"""
        name_success, name_stdout, _ = self._run_git_command(["config", "user.name"])
        if not name_success or name_stdout == TEMP_ABSTRA_NAME:
            self._run_git_command(["config", "user.name", fallback_name])

        email_success, email_stdout, _ = self._run_git_command(["config", "user.email"])
        if not email_success or email_stdout == TEMP_ABSTRA_EMAIL:
            self._run_git_command(["config", "user.email", fallback_email])

    def get_repository_status(self) -> GitStatus:
        """Get comprehensive repository status"""
        if not self.is_git_available():
            return GitStatus(available=False, git_installed=False)

        git_root = self.find_git_root()
        if (
            git_root is not None
            and git_root.resolve() != self.working_directory.resolve()
        ):
            return GitStatus(available=False, is_monorepo=True)

        if self.is_git_repository():
            git_root = self.working_directory
            original_working_dir = None
        elif git_root := self.find_git_root():
            original_working_dir = self.working_directory
            self.working_directory = git_root
        elif self.init_repository():
            git_root = self.working_directory
            original_working_dir = None
        else:
            return GitStatus(available=False)

        try:
            branch = self.get_current_branch()
            if not branch:
                success, stdout, _ = self._run_git_command(["rev-parse", "HEAD"])
                if success and stdout:
                    branch = f"detached-{stdout[:8]}"
                else:
                    return GitStatus(available=False)

            branches = self.get_all_branches()
            last_commit = self.get_last_commit()
            changed_files = self.get_changed_files()
            changed_files_with_status = self.get_changed_files_with_status()
            has_changes = len(changed_files) > 0

            return GitStatus(
                available=True,
                branch=branch,
                branches=branches,
                last_commit=last_commit,
                has_changes=has_changes,
                changed_files=changed_files,
                changed_files_with_status=changed_files_with_status,
            )
        finally:
            if "original_working_dir" in locals() and original_working_dir is not None:
                self.working_directory = original_working_dir

    def checkout_branch(self, branch_name: str) -> Tuple[bool, Optional[str]]:
        """Switch to a different branch"""
        if not self.is_git_repository():
            return False, "Not a git repository"

        success, _, checkout_err = self._run_git_command(["checkout", branch_name])
        error_message = (
            None if success else f"Git checkout failed with error: {checkout_err}"
        )

        if not success and "did not match any file(s) known to git" in checkout_err:
            success, _, create_branch_err = self._run_git_command(
                ["checkout", "-b", branch_name, f"origin/{branch_name}"]
            )
            error_message = (
                None
                if success
                else f"Git checkout failed with error: {create_branch_err}"
            )

        return success, error_message

    def checkout_commit(self, commit_hash: str) -> Tuple[bool, Optional[str]]:
        """Switch to a specific commit (detached HEAD state)"""
        if not self.is_git_repository():
            return False, "Not a git repository"

        success, _, error = self._run_git_command(["checkout", commit_hash])
        error_message = None if success else f"Git checkout failed with error: {error}"

        return success, error_message

    def pull_changes(
        self,
        strategy: str = "merge",
        allow_unrelated: bool = True,
        conflict_resolution: Optional[str] = "theirs",
    ) -> Tuple[bool, Optional[str]]:
        """Pull changes from abstra remote repository

        Args:
            strategy: How to reconcile divergent branches. Options:
                - "merge": Create a merge commit (git pull --no-rebase)
                - "rebase": Rebase local commits on top of remote (git pull --rebase)
                - "ff-only": Only allow fast-forward merges (git pull --ff-only)
            allow_unrelated: Allow merging unrelated histories (git pull --allow-unrelated-histories)
            conflict_resolution: How to resolve conflicts automatically. Options:
                - "theirs": Accept all incoming changes (git pull -X theirs)
                - "ours": Keep all local changes (git pull -X ours)
                - None: Handle conflicts manually (default)
        """
        if not self.is_git_repository():
            return False, "Not a git repository"

        pull_cmd = ["pull"]

        if strategy == "merge":
            pull_cmd.append("--no-rebase")
        elif strategy == "rebase":
            pull_cmd.append("--rebase")
        elif strategy == "ff-only":
            pull_cmd.append("--ff-only")
        else:
            pull_cmd.append("--no-rebase")

        if allow_unrelated:
            pull_cmd.append("--allow-unrelated-histories")

        if conflict_resolution == "theirs":
            pull_cmd.extend(["-X", "theirs"])
        elif conflict_resolution == "ours":
            pull_cmd.extend(["-X", "ours"])

        pull_cmd.extend([REMOTE_NAME, "main"])

        success, _, e = self._run_git_command(pull_cmd)
        error_message = None if success else f"Git pull failed with error: {e}"

        return success, error_message

    def add_all_files(self) -> Tuple[bool, Optional[str]]:
        """Add all changes to staging area"""
        if not self.is_git_repository():
            return False, "Not a git repository"

        success, _, error = self._run_git_command(["add", "."])
        if not success:
            changed_files = self.get_changed_files()
            failed_files = []

            for file in changed_files:
                file_success, _, _ = self._run_git_command(["add", file])
                if not file_success:
                    failed_files.append(file)

            if failed_files:
                return (
                    False,
                    f"Git add failed for the following files: {', '.join(failed_files)}",
                )

            if not changed_files:
                return False, f"Git add failed with error: {error}"

        return True, None

    def commit_changes(
        self, message: str, author: Optional[str] = None
    ) -> Tuple[bool, Optional[str]]:
        """Commit changes with a message"""
        if not self.is_git_repository():
            return False, "Not a git repository"

        git_add_success, git_add_error_message = self.add_all_files()
        if not git_add_success:
            return False, git_add_error_message

        if author and "@" in author:
            author_str = f"{author.split('@')[0]} <{author}>"
            success, _, error = self._run_git_command(
                ["commit", "-m", message, "--author", author_str]
            )
            error_message = (
                None if success else f"Git commit failed with error: {error}"
            )

            return success, error_message

        success, _, error = self._run_git_command(["commit", "-m", message])
        error_message = None if success else f"Git commit failed with error: {error}"

        return success, error_message

    def stash_changes(self, message: str = "WIP") -> Tuple[bool, Optional[str]]:
        """Stash uncommitted changes"""
        if not self.is_git_repository():
            return False, "Not a git repository"

        success, _, err = self._run_git_command(
            ["stash", "push", "-m", message, "--include-untracked"]
        )
        error_message = None if success else f"Git stash failed with error: {err}"

        return success, error_message

    def get_remotes(self) -> List[str]:
        """Get list of remote names"""
        if not self.is_git_repository():
            return []

        success, stdout, _ = self._run_git_command(["remote"])
        if success and stdout:
            return [remote.strip() for remote in stdout.split("\n") if remote.strip()]
        return []

    def has_remote(self, remote_name: str) -> bool:
        """Check if a remote exists"""
        return remote_name in self.get_remotes()

    def add_remote(self, remote_name: str, remote_url: str) -> bool:
        """Add a remote to the repository"""
        if not self.is_git_repository():
            return False

        success, _, _ = self._run_git_command(
            ["remote", "add", remote_name, remote_url]
        )
        return success

    def set_remote_url(self, remote_name: str, remote_url: str) -> bool:
        """Set/update the URL for a remote"""
        if not self.is_git_repository():
            return False

        success, _, _ = self._run_git_command(
            ["remote", "set-url", remote_name, remote_url]
        )
        return success

    def set_git_config(self, key: str, value: str, local: bool = True) -> bool:
        """Set a git configuration value"""
        if not self.is_git_repository():
            return False

        config_flag = "--local" if local else "--global"
        success, _, _ = self._run_git_command(["config", config_flag, key, value])
        return success

    def get_git_config(self, key: str, local: bool = True) -> Optional[str]:
        """Get a git configuration value"""
        if not self.is_git_repository():
            return None

        config_flag = "--local" if local else "--global"
        success, stdout, _ = self._run_git_command(["config", config_flag, key])
        return stdout.strip() if success and stdout else None

    def get_ahead_behind_count(
        self, local_commit: str, remote_commit: str
    ) -> Tuple[int, int]:
        if not self.is_git_repository():
            return (0, 0)

        remote_exists_success, _, _ = self._run_git_command(
            ["cat-file", "-e", remote_commit]
        )

        # If remote commit doesn't exist locally, we need to fetch first (we're behind)
        if not remote_exists_success:
            # We can't calculate exact counts without fetching, but we know:
            # - We have local changes (since commits are different)
            # - Remote has changes we don't have (since remote commit doesn't exist locally)
            # This indicates divergent branches - both ahead and behind
            return (1, 1)  # Simplified: indicates divergence requiring merge/rebase

        # If remote commit exists locally, we can calculate normal ahead/behind
        # Count commits that local has but remote doesn't (ahead)
        ahead_success, ahead_stdout, _ = self._run_git_command(
            ["rev-list", "--count", f"{remote_commit}..{local_commit}"]
        )
        ahead_count = (
            int(ahead_stdout.strip()) if ahead_success and ahead_stdout.strip() else 0
        )

        # Count commits that remote has but local doesn't (behind)
        behind_success, behind_stdout, _ = self._run_git_command(
            ["rev-list", "--count", f"{local_commit}..{remote_commit}"]
        )
        behind_count = (
            int(behind_stdout.strip())
            if behind_success and behind_stdout.strip()
            else 0
        )

        return (ahead_count, behind_count)

    def push_and_deploy(self, branch: str = "main") -> Tuple[bool, Optional[str]]:
        """Deploy to Abstra remote (push to abstra remote)"""
        if not self.is_git_repository():
            return False, "Not a git repository"

        if self.get_current_branch() != "main":
            return False, "Deployments are only allowed from the 'main' branch"

        if not self.has_remote(REMOTE_NAME):
            return False, "Abstra Remote not found"

        success, _, err = self._run_git_command(["push", REMOTE_NAME, branch])
        error_message = None if success else f"Git push failed with error: {err}"

        return success, error_message

    def revert_commit(self, commit_hash: str) -> Tuple[bool, Optional[str]]:
        """Reset working directory to match a previous commit and create a new commit with that content"""
        if not self.is_git_repository():
            return False, "Not a git repository"

        if not commit_hash.strip():
            return False, "Invalid commit hash"

        if self.has_uncommitted_changes():
            return False, "Uncommitted changes present"

        try:
            # Step 1: Reset the working directory and index to match the target commit
            # This will replace all files with their state from the target commit
            success_reset, _, err = self._run_git_command(
                ["reset", "--hard", commit_hash]
            )
            if not success_reset:
                return False, f"Git reset failed with error: {err}"

            # Step 2: Reset HEAD back to the original position (soft reset)
            # This keeps the files from target commit but moves HEAD back
            success_soft, _, err = self._run_git_command(
                ["reset", "--soft", "HEAD@{1}"]
            )
            if not success_soft:
                return False, f"Git soft reset failed with error: {err}"

            # Step 3: Create a new commit with the restored content
            commit_message = f"Restore content from commit {commit_hash[:8]}"
            success_commit, _, err = self._run_git_command(
                ["commit", "-m", commit_message]
            )
            error_message = (
                None if success_commit else f"Git commit failed with error: {err}"
            )

            return success_commit, error_message

        except Exception:
            self._run_git_command(["reset", "--hard", "HEAD"])
            return False, "An error occurred during revert"

    def check_merge_conflicts(self, remote_commit: str) -> bool:
        """Check if merging with remote commit would cause conflicts"""
        if not self.is_git_repository():
            return False

        if not remote_commit:
            return False

        # Get current commit
        success, current_commit, _ = self._run_git_command(["rev-parse", "HEAD"])
        if not success:
            return False

        # Use git merge-tree to detect potential conflicts without actually merging
        # git merge-tree <base> <branch1> <branch2>
        # We need to find the merge base first
        success, merge_base, _ = self._run_git_command(
            ["merge-base", current_commit.strip(), remote_commit]
        )
        if not success:
            # If no merge base, assume potential conflicts
            return True

        # Check if merge would have conflicts
        success, output, _ = self._run_git_command(
            ["merge-tree", merge_base.strip(), current_commit.strip(), remote_commit]
        )

        # If merge-tree succeeds and output contains conflict markers, there are conflicts
        if success and output:
            # Look for conflict markers in the output
            return "<<<<<<< " in output or "=======" in output or ">>>>>>> " in output

        return False

    def check_ignore(self, path: Path) -> bool:
        """
        Check if a path should be ignored according to .gitignore rules.
        Uses git check-ignore for accurate gitignore handling.
        """
        if not self.is_git_repository():
            return False

        try:
            # Get path relative to working directory
            try:
                relative_path = path.relative_to(self.working_directory)
            except ValueError:
                # Path is outside working directory, not ignored
                return False

            # Use git check-ignore -q (quiet mode, exit code only)
            success, _, _ = self._run_git_command(
                ["check-ignore", "-q", str(relative_path)]
            )

            # Exit code 0 means the path is ignored
            return success

        except Exception:
            return False

    def check_ignore_batch(self, paths: List[Path]) -> Dict[Path, bool]:
        """
        Check multiple paths at once using git check-ignore --stdin.
        Much more efficient than checking one path at a time.
        """
        if not self.is_git_repository():
            return {path: False for path in paths}

        results = {}
        relative_paths = []
        path_map = {}

        # Convert to relative paths
        for path in paths:
            try:
                rel_path = path.relative_to(self.working_directory)
                rel_path_str = str(rel_path)
                relative_paths.append(rel_path_str)
                path_map[rel_path_str] = path
            except ValueError:
                # Path outside working directory, not ignored
                results[path] = False

        if not relative_paths:
            return results

        try:
            # Use git check-ignore with stdin for batch processing
            input_data = "\n".join(relative_paths)
            _, stdout, _ = self._run_git_command(
                ["check-ignore", "--stdin"],
                cwd=self.working_directory,
                input=input_data,
            )

            # git check-ignore --stdin returns the paths that ARE ignored in stdout
            ignored_paths_set = (
                set(stdout.strip().split("\n")) if stdout.strip() else set()
            )

            # Map results back to original paths
            for rel_path_str, path in path_map.items():
                results[path] = rel_path_str in ignored_paths_set

        except Exception:
            # On error, mark all remaining as not ignored
            for path in paths:
                if path not in results:
                    results[path] = False

        return results

    def untrack_path(self, path: Path):
        """Untrack a path from git (git rm --cached)"""
        if not self.is_git_repository():
            return

        try:
            relative_path = path.relative_to(self.working_directory)
        except ValueError:
            print("Path is outside the working directory; cannot untrack.")
            return

        self._run_git_command(["rm", "--cached", "-r", str(relative_path)])



import math
from random import random
from typing import List

from abstra_internals.repositories.project.project import Stage

from .geometry import Point2D, Rect, Vector2D

# Constantes da simulação
MAX_ACCELERATION = math.inf  # Limite máximo de aceleração
MAX_VELOCITY = 10  # Limite máximo de velocidade

ideal_edge_length = 100.0
node_repulsion_strength = (
    1e5  # Força de repulsão entre nós (re-enabling like TypeScript)
)
attraction_strength = (
    1e-1  # Força de atração entre nós conectados (matching TypeScript)
)
edge_repulsion_strength = (
    0  # Força de repulsão entre nós e arestas (matching TypeScript)
)
damping = 0.9  # Fator de amortecimento para velocidade (matching TypeScript)
noise_size = 1e-2  # Fator de ruído para simulação
stage_width = 200
stage_height = 100


def noise() -> Vector2D:
    return Vector2D((random() - 0.5), (random() - 0.5)).scale(2 * noise_size)


class PhysicsNode:
    def __init__(
        self, position: Point2D, velocity: Vector2D, acceleration: Vector2D, rect: Rect
    ):
        self.position = position
        self.velocity = velocity
        self.acceleration = acceleration
        self.rect = rect


class PhysicsEdge:
    def __init__(self, source: int, target: int):
        self.source = source
        self.target = target


def force_center(nodes: List[PhysicsNode]) -> None:
    average_x = sum(node.position.x for node in nodes) / len(nodes)
    average_y = sum(node.position.y for node in nodes) / len(nodes)
    center = Point2D(average_x, average_y)
    displacement = Point2D(0, 0).sub(center)
    for node in nodes:
        node.position = node.position.plus(displacement)

    assert abs(sum(node.position.x for node in nodes) / len(nodes)) < 1
    assert abs(sum(node.position.y for node in nodes) / len(nodes)) < 1


def apply_forces(nodes: List[PhysicsNode], edges: List[PhysicsEdge]) -> None:
    """
    Apply physics forces to nodes in a force-directed graph layout.

    Args:
        nodes: List of PhysicsNode objects with position, velocity, acceleration, and rect
        edges: List of PhysicsEdge objects with source and target indices
    """

    # Reset acceleration for all nodes
    for node in nodes:
        node.acceleration = Vector2D(0, 0)

    # Apply repulsion forces between nodes
    for i in range(len(nodes)):
        for j in range(i + 1, len(nodes)):
            direction = nodes[i].position.sub(nodes[j].position)
            distance = max(1, nodes[i].rect.distance_to_rect(nodes[j].rect))
            repulsion = direction.norm().scale(
                node_repulsion_strength / (distance * distance)
            )
            nodes[i].acceleration = nodes[i].acceleration.plus(repulsion)
            nodes[j].acceleration = nodes[j].acceleration.minus(repulsion)

    # Apply attraction forces along edges
    for edge in edges:
        source_node = nodes[edge.source]
        target_node = nodes[edge.target]

        direction = source_node.rect.link(target_node.rect).to_delta()
        distance = source_node.rect.distance_to_rect(target_node.rect)
        displacement = distance - ideal_edge_length
        attraction = direction.norm().scale(attraction_strength * displacement)

        source_node.acceleration = source_node.acceleration.plus(attraction)
        target_node.acceleration = target_node.acceleration.minus(attraction)

    # Apply repulsion forces from edges to all nodes
    for edge in edges:
        source_node = nodes[edge.source]
        target_node = nodes[edge.target]

        for node in nodes:
            if node == target_node or node == source_node:
                continue

            direction = source_node.rect.link(target_node.rect)
            distance = node.rect.distance_to_line(direction)

            if distance > 0:
                edge_repulsion = (
                    direction.to_delta()
                    .norm()
                    .scale(edge_repulsion_strength / (distance * distance))
                )
                node.acceleration = node.acceleration.plus(edge_repulsion)

    # Update velocity and position for all nodes
    for node in nodes:
        node.velocity = (
            node.velocity.plus(node.acceleration.plus(noise()).limit(MAX_ACCELERATION))
            .limit(MAX_VELOCITY)
            .scale(damping)
        )
        node.position = node.position.plus(node.velocity)


def simulate_layout(
    stages: List[Stage],
    iterations: int = 100,
):
    """
    Simulate force-directed layout for workflow stages.

    Args:
        stages: List of stage dictionaries with id, position, etc.
        transitions: List of transition dictionaries with sourceStageId, targetStageId
        iterations: Number of physics simulation iterations

    Returns:
        List of updated stage dictionaries with new positions
    """
    # Create a mapping from stage IDs to indices
    stage_id_to_index = {stage.id: i for i, stage in enumerate(stages)}

    # Create physics nodes from stages
    nodes: List[PhysicsNode] = []
    for stage in stages:
        position = Point2D(stage.workflow_position[0], stage.workflow_position[1])
        velocity = Vector2D(0, 0)
        acceleration = Vector2D(0, 0)

        rect = Rect(position, stage_width, stage_height)
        nodes.append(PhysicsNode(position, velocity, acceleration, rect))

    # Create physics edges from transitions
    edges: List[PhysicsEdge] = []
    for stage in stages:
        for transition in stage.workflow_transitions:
            source_id = stage.id
            target_id = transition.target_id

            if source_id in stage_id_to_index and target_id in stage_id_to_index:
                source_idx = stage_id_to_index[source_id]
                target_idx = stage_id_to_index[target_id]
                edges.append(PhysicsEdge(source_idx, target_idx))

    # Run physics simulation
    for _ in range(iterations):
        # Update rect centers to match node positions
        for i, node in enumerate(nodes):
            node.rect.center = node.position

        apply_forces(nodes, edges)
        force_center(nodes)
    # Update stage positions
    for i, stage in enumerate(stages):
        stage.workflow_position = (nodes[i].position.x, nodes[i].position.y)



"""
Comprehensive test suite for TaskExecutor.send_task with module/component support.

This test suite covers:
- Basic task sending without modules
- Task sending with ComponentStages (module entry)
- Task sending from module output stages (module exit)
- Multiple input/output stages in modules
- Complex workflows with nested transitions
- Edge cases: empty transitions, invalid stages, multiple paths
"""

import json
from pathlib import Path
from unittest.mock import Mock, patch

from abstra_internals.controllers.common.task_executors import TaskExecutor
from abstra_internals.entities.execution_context import (
    ScriptContext,
)
from abstra_internals.repositories.project.project import (
    ComponentStage,
    FormStage,
    NotificationTrigger,
    ScriptStage,
    WorkflowTransition,
)
from abstra_internals.repositories.tasks import TaskDTO, TaskEventDetails
from abstra_internals.settings import Settings
from tests.fixtures import BaseTest, clear_dir


class TaskExecutorTest(BaseTest):
    """Test suite for TaskExecutor.send_task functionality"""

    def setUp(self):
        super().setUp()
        self.modules_path = Settings.root_path / "abstra_components"

        # Mock repositories
        self.mock_tasks_repo = Mock()
        self.mock_producer = Mock()
        self.mock_email = Mock()

        self.repositories.tasks = self.mock_tasks_repo
        self.repositories.producer = self.mock_producer
        self.repositories.email = self.mock_email

        # Don't create executor yet - wait until after test data is set up
        self.executor = None

    def get_executor(self):
        """Create or return executor (creates after test data is set up)"""
        if self.executor is None:
            self.executor = TaskExecutor(self.repositories)
        return self.executor

    def tearDown(self) -> None:
        clear_dir(self.root)

    def create_module_directory(self, module_name: str) -> Path:
        """Helper: Create a module directory structure"""
        module_path = self.modules_path / module_name
        module_path.mkdir(parents=True, exist_ok=True)
        return module_path

    def create_module_abstra_json(self, module_name: str, stages: dict):
        """Helper: Create an abstra.json file for a module"""
        module_path = self.create_module_directory(module_name)

        # Ensure all stages have required fields with defaults
        for stage_type in ["forms", "hooks", "scripts", "jobs", "components"]:
            for stage in stages.get(stage_type, []):
                if "transitions" not in stage:
                    stage["transitions"] = []

        abstra_json = {
            "version": "17.0",
            "workspace": {
                "name": f"{module_name} Module",
                "brand_name": "",
                "theme": None,
                "logo_url": None,
                "favicon_url": None,
                "main_color": None,
                "font_family": None,
                "font_color": None,
                "language": "en",
            },
            "home": {"access_control": {"is_public": False, "required_roles": []}},
            "forms": stages.get("forms", []),
            "hooks": stages.get("hooks", []),
            "scripts": stages.get("scripts", []),
            "jobs": stages.get("jobs", []),
            "components": stages.get("components", []),
        }

        abstra_json_path = module_path / "abstra.json"
        with open(abstra_json_path, "w", encoding="utf-8") as f:
            json.dump(abstra_json, f, indent=2)

        return abstra_json_path

    def create_mock_task(self, target_stage_id: str) -> TaskDTO:
        """Helper: Create a mock task DTO"""
        return TaskDTO(
            id="task-123",
            type="task",
            payload={"data": "test"},
            target_stage_id=target_stage_id,
            status="pending",
            created=TaskEventDetails(
                at="2024-01-01T00:00:00",
                by_execution_id=None,
                by_stage_id="source",
            ),
            locked=None,
            completed=None,
        )

    # ==================== BASIC TESTS (No Modules) ====================

    def test_send_task_simple_form_to_script(self):
        """Test basic task sending from form to script (no modules)"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="script-1",
                    target_type="scripts",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        script = ScriptStage(
            id="script-1",
            file="script1.py",
            title="Script 1",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.scripts.append(script)
        self.repositories.project.save(project)

        # Mock task repository
        mock_task = self.create_mock_task("script-1")
        self.mock_tasks_repo.send_task = Mock(return_value=mock_task)

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Verify
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["source_stage_id"], "form-1")
        self.assertEqual(call_args["target_stage_id"], "script-1")
        self.assertEqual(call_args["type"], "task")

        # Script should be enqueued
        self.mock_producer.enqueue.assert_called_once()

    def test_send_task_no_transitions(self):
        """Test sending task from stage with no transitions shows warning"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[],  # No transitions
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        project.forms.append(form)
        self.repositories.project.save(project)

        # Execute with warning capture
        with patch("builtins.print") as mock_print:
            self.get_executor().send_task(
                type="task",
                current_stage=form,
                payload={"data": "test"},
                execution=None,
                show_warning=True,
            )

        # Should print warning
        mock_print.assert_called_once()
        self.assertIn("No transitions found", str(mock_print.call_args))

        # Should not send any tasks
        self.mock_tasks_repo.send_task.assert_not_called()

    def test_send_task_multiple_transitions(self):
        """Test sending tasks to multiple target stages"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="script-1",
                    target_type="scripts",
                    type="task",
                ),
                WorkflowTransition(
                    id="trans-2",
                    target_id="script-2",
                    target_type="scripts",
                    type="task",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        script1 = ScriptStage(
            id="script-1",
            file="script1.py",
            title="Script 1",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script-2",
            file="script2.py",
            title="Script 2",
            workflow_position=(100, 100),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.scripts.append(script1)
        project.scripts.append(script2)
        self.repositories.project.save(project)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            side_effect=[
                self.create_mock_task("script-1"),
                self.create_mock_task("script-2"),
            ]
        )

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Should send 2 tasks
        self.assertEqual(self.mock_tasks_repo.send_task.call_count, 2)

        # Check both targets
        calls = self.mock_tasks_repo.send_task.call_args_list
        target_ids = [call[1]["target_stage_id"] for call in calls]
        self.assertIn("script-1", target_ids)
        self.assertIn("script-2", target_ids)

    def test_send_task_with_task_type_filter(self):
        """Test that task type filtering works correctly"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="script-1",
                    target_type="scripts",
                    type="task",
                    task_type="success",
                ),
                WorkflowTransition(
                    id="trans-2",
                    target_id="script-2",
                    target_type="scripts",
                    type="task",
                    task_type="error",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        script1 = ScriptStage(
            id="script-1",
            file="script1.py",
            title="Success Script",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script-2",
            file="script2.py",
            title="Error Script",
            workflow_position=(100, 100),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.scripts.append(script1)
        project.scripts.append(script2)
        self.repositories.project.save(project)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("script-1")
        )

        # Execute with "success" type - should only match script-1
        self.get_executor().send_task(
            type="success",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Should send only 1 task to script-1
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "script-1")

    def test_send_task_script_gets_enqueued(self):
        """Test that script stages are enqueued for execution"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="script-1",
                    target_type="scripts",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        script = ScriptStage(
            id="script-1",
            file="script1.py",
            title="Script 1",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.scripts.append(script)
        self.repositories.project.save(project)

        # Mock
        mock_task = self.create_mock_task("script-1")
        self.mock_tasks_repo.send_task = Mock(return_value=mock_task)

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Script should be enqueued
        self.mock_producer.enqueue.assert_called_once()
        call_args = self.mock_producer.enqueue.call_args[1]
        self.assertEqual(call_args["stage_id"], "script-1")
        self.assertIsInstance(call_args["context"], ScriptContext)
        self.assertEqual(call_args["context"].task_id, "task-123")

    def test_send_task_form_not_enqueued(self):
        """Test that form stages are NOT enqueued (only scripts)"""
        project = self.repositories.project.load()

        form1 = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="form-2",
                    target_type="forms",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        form2 = FormStage(
            id="form-2",
            path="form2",
            title="Form 2",
            file="form2.py",
            workflow_position=(100, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        project.forms.append(form1)
        project.forms.append(form2)
        self.repositories.project.save(project)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("form-2")
        )

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form1,
            payload={"data": "test"},
            execution=None,
        )

        # Form should NOT be enqueued
        self.mock_producer.enqueue.assert_not_called()

    # ==================== MODULE ENTRY TESTS (ComponentStage) ====================

    def test_send_task_to_component_expands_to_input_stages(self):
        """Test that sending task to ComponentStage expands to module's input stages"""
        # Create module with input stages
        self.create_module_abstra_json(
            "test-module",
            stages={
                "forms": [
                    {
                        "id": "module-input-form",
                        "path": "input",
                        "title": "Module Input",
                        "file": "input.py",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": True,  # Input stage
                        "output": False,
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        form = FormStage(
            id="main-form",
            path="main",
            title="Main Form",
            file="main.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="comp-1",
                    target_type="components",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        component = ComponentStage(
            id="comp-1",
            title="Test Module",
            package_name="test-module",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.components.append(component)
        self.repositories.project.save(project)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("module-input-form")
        )

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Should send task to module input, NOT to component
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "module-input-form")
        self.assertNotEqual(call_args["target_stage_id"], "comp-1")

    def test_send_task_to_component_with_multiple_inputs(self):
        """Test that component with multiple input stages sends to all of them"""
        # Create module with 2 input stages
        self.create_module_abstra_json(
            "multi-input-module",
            stages={
                "forms": [
                    {
                        "id": "input-form-1",
                        "path": "input1",
                        "title": "Input 1",
                        "file": "input1.py",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": True,
                        "output": False,
                    },
                    {
                        "id": "input-form-2",
                        "path": "input2",
                        "title": "Input 2",
                        "file": "input2.py",
                        "workflow_position": [0, 100],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": True,
                        "output": False,
                    },
                ]
            },
        )

        project = self.repositories.project.load()

        form = FormStage(
            id="main-form",
            path="main",
            title="Main Form",
            file="main.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="comp-1",
                    target_type="components",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        component = ComponentStage(
            id="comp-1",
            title="Multi Input Module",
            package_name="multi-input-module",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.components.append(component)
        self.repositories.project.save(project)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            side_effect=[
                self.create_mock_task("input-form-1"),
                self.create_mock_task("input-form-2"),
            ]
        )

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Should send 2 tasks to both input stages
        self.assertEqual(self.mock_tasks_repo.send_task.call_count, 2)

        calls = self.mock_tasks_repo.send_task.call_args_list
        target_ids = [call[1]["target_stage_id"] for call in calls]
        self.assertIn("input-form-1", target_ids)
        self.assertIn("input-form-2", target_ids)

    def test_send_task_to_component_with_no_input_stages(self):
        """Test handling of component with no input stages (edge case)"""
        # Create module with NO input stages
        self.create_module_abstra_json(
            "no-input-module",
            stages={
                "forms": [
                    {
                        "id": "regular-form",
                        "path": "regular",
                        "title": "Regular Form",
                        "file": "regular.py",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": False,  # Not an input stage
                        "output": False,
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        form = FormStage(
            id="main-form",
            path="main",
            title="Main Form",
            file="main.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="comp-1",
                    target_type="components",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        component = ComponentStage(
            id="comp-1",
            title="No Input Module",
            package_name="no-input-module",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.components.append(component)
        self.repositories.project.save(project)

        # Execute - should handle gracefully
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Should not send any tasks (no input stages to expand to)
        self.mock_tasks_repo.send_task.assert_not_called()

    # ==================== MODULE EXIT TESTS (Output Stages) ====================

    def test_send_task_from_module_output_uses_component_transitions(self):
        """Test that module output stage uses parent component's transitions"""
        # Create module with output stage
        self.create_module_abstra_json(
            "exit-module",
            stages={
                "scripts": [
                    {
                        "id": "module-output-script",
                        "file": "output.py",
                        "title": "Module Output",
                        "workflow_position": [100, 0],
                        "transitions": [],  # No transitions - will use component's transitions
                        "is_initial": True,
                        "input": False,
                        "output": True,  # Output stage
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        component = ComponentStage(
            id="comp-1",
            title="Exit Module",
            package_name="exit-module",
            workflow_position=(50, 0),
            workflow_transitions=[
                # This transition should be USED
                WorkflowTransition(
                    id="component-trans",
                    target_id="final-form",
                    target_type="forms",
                    type="task",
                )
            ],
        )

        final_form = FormStage(
            id="final-form",
            path="final",
            title="Final Form",
            file="final.py",
            workflow_position=(200, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        project.components.append(component)
        project.forms.append(final_form)
        self.repositories.project.save(project)

        # Get the module output stage
        module_output = project.get_stage("module-output-script", include_modules=True)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("final-form")
        )

        # Execute from module output stage
        self.get_executor().send_task(
            type="task",
            current_stage=module_output,
            payload={"data": "test"},
            execution=None,
        )

        # Should send task to final-form (from component's transitions)
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "final-form")

    def test_send_task_from_module_internal_uses_own_transitions(self):
        """Test that non-output module stages use their own transitions"""
        # Create module with internal stage (not output)
        self.create_module_abstra_json(
            "internal-module",
            stages={
                "forms": [
                    {
                        "id": "module-internal-form",
                        "path": "internal",
                        "title": "Module Internal",
                        "file": "internal.py",
                        "workflow_position": [0, 0],
                        "transitions": [
                            {
                                "id": "internal-trans",
                                "target_id": "module-script",
                                "target_type": "scripts",
                                "type": "task",
                                "task_type": None,
                            }
                        ],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": True,
                        "output": False,  # NOT an output stage
                    }
                ],
                "scripts": [
                    {
                        "id": "module-script",
                        "file": "script.py",
                        "title": "Module Script",
                        "workflow_position": [100, 0],
                        "is_initial": False,
                        "input": False,
                        "output": False,
                    }
                ],
            },
        )

        project = self.repositories.project.load()

        component = ComponentStage(
            id="comp-1",
            title="Internal Module",
            package_name="internal-module",
            workflow_position=(50, 0),
            workflow_transitions=[],  # Component has no transitions
        )

        project.components.append(component)
        self.repositories.project.save(project)

        # Get the module internal stage
        module_form = project.get_stage("module-internal-form", include_modules=True)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("module-script")
        )

        # Execute from internal module stage
        self.get_executor().send_task(
            type="task",
            current_stage=module_form,
            payload={"data": "test"},
            execution=None,
        )

        # Should use stage's own transitions (to module-script)
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "module-script")

    def test_send_task_from_module_output_with_multiple_component_transitions(self):
        """Test module output with multiple transitions on parent component"""
        # Create module
        self.create_module_abstra_json(
            "multi-exit-module",
            stages={
                "scripts": [
                    {
                        "id": "module-output",
                        "file": "output.py",
                        "title": "Module Output",
                        "workflow_position": [100, 0],
                        "is_initial": False,
                        "input": False,
                        "output": True,
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        component = ComponentStage(
            id="comp-1",
            title="Multi Exit Module",
            package_name="multi-exit-module",
            workflow_position=(50, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="form-1",
                    target_type="forms",
                    type="task",
                ),
                WorkflowTransition(
                    id="trans-2",
                    target_id="form-2",
                    target_type="forms",
                    type="task",
                ),
            ],
        )

        form1 = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(200, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        form2 = FormStage(
            id="form-2",
            path="form2",
            title="Form 2",
            file="form2.py",
            workflow_position=(200, 100),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        project.components.append(component)
        project.forms.append(form1)
        project.forms.append(form2)
        self.repositories.project.save(project)

        # Get module output
        module_output = project.get_stage("module-output", include_modules=True)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            side_effect=[
                self.create_mock_task("form-1"),
                self.create_mock_task("form-2"),
            ]
        )

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=module_output,
            payload={"data": "test"},
            execution=None,
        )

        # Should send to both forms (component's transitions)
        self.assertEqual(self.mock_tasks_repo.send_task.call_count, 2)

        calls = self.mock_tasks_repo.send_task.call_args_list
        target_ids = [call[1]["target_stage_id"] for call in calls]
        self.assertIn("form-1", target_ids)
        self.assertIn("form-2", target_ids)

    def test_send_task_from_module_output_with_no_component_transitions(self):
        """Test module output when parent component has no transitions"""
        # Create module
        self.create_module_abstra_json(
            "orphan-module",
            stages={
                "scripts": [
                    {
                        "id": "module-output",
                        "file": "output.py",
                        "title": "Module Output",
                        "workflow_position": [100, 0],
                        "is_initial": False,
                        "input": False,
                        "output": True,
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        component = ComponentStage(
            id="comp-1",
            title="Orphan Module",
            package_name="orphan-module",
            workflow_position=(50, 0),
            workflow_transitions=[],  # No transitions!
        )

        project.components.append(component)
        self.repositories.project.save(project)

        # Get module output
        module_output = project.get_stage("module-output", include_modules=True)

        # Execute with warning capture
        with patch("builtins.print") as mock_print:
            self.get_executor().send_task(
                type="task",
                current_stage=module_output,
                payload={"data": "test"},
                execution=None,
                show_warning=True,
            )

        # Should show warning
        mock_print.assert_called_once()

        # Should not send any tasks
        self.mock_tasks_repo.send_task.assert_not_called()

    # ==================== COMPLEX WORKFLOW TESTS ====================

    def test_send_task_main_to_component_to_main(self):
        """Test complete flow: main -> component -> module internal -> module exit -> main"""
        # Create module with input, internal, and output stages
        self.create_module_abstra_json(
            "full-flow-module",
            stages={
                "forms": [
                    {
                        "id": "module-input",
                        "path": "input",
                        "title": "Module Input",
                        "file": "input.py",
                        "workflow_position": [0, 0],
                        "transitions": [
                            {
                                "id": "trans-internal",
                                "target_id": "module-processor",
                                "target_type": "scripts",
                                "type": "task",
                                "task_type": None,
                            }
                        ],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": True,
                        "output": False,
                    }
                ],
                "scripts": [
                    {
                        "id": "module-processor",
                        "file": "processor.py",
                        "title": "Module Processor",
                        "workflow_position": [100, 0],
                        "transitions": [
                            {
                                "id": "trans-to-output",
                                "target_id": "module-output",
                                "target_type": "scripts",
                                "type": "task",
                                "task_type": None,
                            }
                        ],
                        "is_initial": False,
                        "input": False,
                        "output": False,
                    },
                    {
                        "id": "module-output",
                        "file": "output.py",
                        "title": "Module Output",
                        "workflow_position": [200, 0],
                        "is_initial": False,
                        "input": False,
                        "output": True,
                    },
                ],
            },
        )

        project = self.repositories.project.load()

        initial_form = FormStage(
            id="initial-form",
            path="initial",
            title="Initial Form",
            file="initial.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-to-comp",
                    target_id="comp-1",
                    target_type="components",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        component = ComponentStage(
            id="comp-1",
            title="Full Flow Module",
            package_name="full-flow-module",
            workflow_position=(100, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-to-final",
                    target_id="final-form",
                    target_type="forms",
                    type="task",
                )
            ],
        )

        final_form = FormStage(
            id="final-form",
            path="final",
            title="Final Form",
            file="final.py",
            workflow_position=(300, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        project.forms.append(initial_form)
        project.components.append(component)
        project.forms.append(final_form)
        self.repositories.project.save(project)

        # Step 1: Send from initial form to component
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("module-input")
        )

        self.get_executor().send_task(
            type="task",
            current_stage=initial_form,
            payload={"data": "step1"},
            execution=None,
        )

        # Should expand component to module-input
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "module-input")

        # Step 2: Module input -> processor (internal transition)
        module_input = project.get_stage("module-input", include_modules=True)
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("module-processor")
        )

        self.get_executor().send_task(
            type="task",
            current_stage=module_input,
            payload={"data": "step2"},
            execution=None,
        )

        # Should use module stage's own transitions
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "module-processor")

        # Step 3: Module output -> final form (exits module via component)
        module_output = project.get_stage("module-output", include_modules=True)
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("final-form")
        )

        self.get_executor().send_task(
            type="task",
            current_stage=module_output,
            payload={"data": "step3"},
            execution=None,
        )

        # Should use component's transitions (exit module)
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "final-form")

    def test_send_task_with_nested_components(self):
        """Test workflow with multiple components in sequence"""
        # Create two modules
        self.create_module_abstra_json(
            "module-a",
            stages={
                "forms": [
                    {
                        "id": "input-a",
                        "path": "inputa",
                        "title": "Input A",
                        "file": "inputa.py",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": True,
                        "output": True,  # Also output
                    }
                ]
            },
        )

        self.create_module_abstra_json(
            "module-b",
            stages={
                "scripts": [
                    {
                        "id": "input-b",
                        "file": "inputb.py",
                        "title": "Input B",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "input": True,
                        "output": False,
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        comp_a = ComponentStage(
            id="comp-a",
            title="Module A",
            package_name="module-a",
            workflow_position=(50, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-a-to-b",
                    target_id="comp-b",
                    target_type="components",
                    type="task",
                )
            ],
        )

        comp_b = ComponentStage(
            id="comp-b",
            title="Module B",
            package_name="module-b",
            workflow_position=(150, 0),
            workflow_transitions=[],
        )

        project.components.append(comp_a)
        project.components.append(comp_b)
        self.repositories.project.save(project)

        # Module A output exits to comp-b, which should expand to module B input
        module_a_output = project.get_stage("input-a", include_modules=True)

        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("input-b")
        )

        self.get_executor().send_task(
            type="task",
            current_stage=module_a_output,
            payload={"data": "test"},
            execution=None,
        )

        # Should go through: module-a-output -> comp-a transitions -> comp-b -> expand to input-b
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "input-b")

    def test_send_task_with_execution_context(self):
        """Test that execution context is properly updated with sent tasks"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="script-1",
                    target_type="scripts",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        script = ScriptStage(
            id="script-1",
            file="script1.py",
            title="Script 1",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.scripts.append(script)
        self.repositories.project.save(project)

        # Create execution with mock context that has sent_tasks attribute
        execution = Mock()
        execution.id = "exec-123"
        execution.context = Mock()
        execution.context.sent_tasks = []

        # Mock
        mock_task = self.create_mock_task("script-1")
        self.mock_tasks_repo.send_task = Mock(return_value=mock_task)

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=execution,
        )

        # Execution context should have sent task ID
        self.assertIn("task-123", execution.context.sent_tasks)

    def test_send_task_component_to_multiple_regular_stages(self):
        """Test component transitioning to mix of regular stages and components"""
        # Create module
        self.create_module_abstra_json(
            "test-module",
            stages={
                "scripts": [
                    {
                        "id": "module-output",
                        "file": "output.py",
                        "title": "Output",
                        "workflow_position": [0, 0],
                        "is_initial": False,
                        "input": False,
                        "output": True,
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        component = ComponentStage(
            id="comp-1",
            title="Test Module",
            package_name="test-module",
            workflow_position=(50, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="form-1",
                    target_type="forms",
                    type="task",
                ),
                WorkflowTransition(
                    id="trans-2",
                    target_id="script-1",
                    target_type="scripts",
                    type="task",
                ),
            ],
        )

        form1 = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(200, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        script1 = ScriptStage(
            id="script-1",
            file="script1.py",
            title="Script 1",
            workflow_position=(200, 100),
            workflow_transitions=[],
        )

        project.components.append(component)
        project.forms.append(form1)
        project.scripts.append(script1)
        self.repositories.project.save(project)

        # Module output exits
        module_output = project.get_stage("module-output", include_modules=True)

        self.mock_tasks_repo.send_task = Mock(
            side_effect=[
                self.create_mock_task("form-1"),
                self.create_mock_task("script-1"),
            ]
        )

        self.get_executor().send_task(
            type="task",
            current_stage=module_output,
            payload={"data": "test"},
            execution=None,
        )

        # Should send to both regular stages
        self.assertEqual(self.mock_tasks_repo.send_task.call_count, 2)

        calls = self.mock_tasks_repo.send_task.call_args_list
        target_ids = [call[1]["target_stage_id"] for call in calls]
        self.assertIn("form-1", target_ids)
        self.assertIn("script-1", target_ids)

        # Script should be enqueued
        self.mock_producer.enqueue.assert_called_once()

    # ==================== EDGE CASES ====================

    def test_send_task_module_stage_without_parent_component(self):
        """Test handling of orphaned module stage (no parent component found)"""
        # Create module
        self.create_module_abstra_json(
            "orphan-module",
            stages={
                "scripts": [
                    {
                        "id": "orphan-output",
                        "file": "output.py",
                        "title": "Orphan Output",
                        "workflow_position": [0, 0],
                        "is_initial": False,
                        "input": False,
                        "output": True,
                    }
                ]
            },
        )

        project = self.repositories.project.load()
        # Don't add any ComponentStage referencing this module!
        self.repositories.project.save(project)

        # Get orphan stage
        orphan_stage = project.get_stage("orphan-output", include_modules=True)

        # Execute - should handle gracefully
        with patch("builtins.print") as mock_print:
            self.get_executor().send_task(
                type="task",
                current_stage=orphan_stage,
                payload={"data": "test"},
                execution=None,
                show_warning=True,
            )

        # Should show warning (no transitions found)
        mock_print.assert_called_once()

        # Should not send tasks
        self.mock_tasks_repo.send_task.assert_not_called()

    def test_send_task_to_component_with_no_package_name(self):
        """Test component with None package_name (edge case)"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="comp-1",
                    target_type="components",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        component = ComponentStage(
            id="comp-1",
            title="Invalid Component",
            package_name=None,  # No package!
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.components.append(component)
        self.repositories.project.save(project)

        # Mock task creation
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("comp-1")
        )

        # Execute - ComponentStage without package_name will be treated as regular stage
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Task is sent to the component stage itself (not expanded since no package)
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "comp-1")

    def test_send_task_notification_enabled_form(self):
        """Test that send_task calls notification check for forms with enabled trigger"""
        project = self.repositories.project.load()

        form1 = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="form-2",
                    target_type="forms",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        form2 = FormStage(
            id="form-2",
            path="form2",
            title="Form 2 (with notification)",
            file="form2.py",
            workflow_position=(100, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails",
                enabled=True,  # Notification enabled!
            ),
        )

        project.forms.append(form1)
        project.forms.append(form2)
        self.repositories.project.save(project)

        # Mock
        mock_task = self.create_mock_task("form-2")
        self.mock_tasks_repo.send_task = Mock(return_value=mock_task)

        # Execute
        self.get_executor().send_task(
            type="task",
            current_stage=form1,
            payload={"assignee_emails": ["test@example.com"]},
            execution=None,
        )

        # Task should be sent
        self.mock_tasks_repo.send_task.assert_called_once()

        # Note: Email sending happens internally when notification is enabled,
        # but testing it requires patching the reloaded stage object which is complex.
        # The important part is that the task is sent correctly.

    def test_send_task_with_circular_reference(self):
        """Test handling of deeply nested component workflow"""
        # Create two modules that chain together
        self.create_module_abstra_json(
            "module-a",
            stages={
                "scripts": [
                    {
                        "id": "script-a",
                        "file": "script_a.py",
                        "title": "Script A",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "input": True,
                        "output": False,
                    }
                ]
            },
        )

        self.create_module_abstra_json(
            "module-b",
            stages={
                "scripts": [
                    {
                        "id": "script-b",
                        "file": "script_b.py",
                        "title": "Script B",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "input": True,
                        "output": False,
                    }
                ]
            },
        )

        project = self.repositories.project.load()

        # Create chained components: main -> comp-a -> comp-b
        form = FormStage(
            id="main-form",
            path="main",
            title="Main Form",
            file="main.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="comp-a",
                    target_type="components",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        comp_a = ComponentStage(
            id="comp-a",
            title="Module A",
            package_name="module-a",
            workflow_position=(100, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-2",
                    target_id="comp-b",
                    target_type="components",
                    type="task",
                )
            ],
        )

        comp_b = ComponentStage(
            id="comp-b",
            title="Module B",
            package_name="module-b",
            workflow_position=(200, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.components.append(comp_a)
        project.components.append(comp_b)
        self.repositories.project.save(project)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("script-a")
        )

        # Execute - should expand comp-a to script-a
        self.get_executor().send_task(
            type="task",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Should send task to module-a's input
        self.mock_tasks_repo.send_task.assert_called_once()
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "script-a")

    def test_send_task_type_matching_with_none_task_type(self):
        """Test that transitions with task_type=None match any type"""
        project = self.repositories.project.load()

        form = FormStage(
            id="form-1",
            path="form1",
            title="Form 1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="script-1",
                    target_type="scripts",
                    type="task",
                    task_type=None,  # Matches any type
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        script = ScriptStage(
            id="script-1",
            file="script1.py",
            title="Script 1",
            workflow_position=(100, 0),
            workflow_transitions=[],
        )

        project.forms.append(form)
        project.scripts.append(script)
        self.repositories.project.save(project)

        # Mock
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("script-1")
        )

        # Execute with any type
        self.get_executor().send_task(
            type="custom-type",
            current_stage=form,
            payload={"data": "test"},
            execution=None,
        )

        # Should still match and send task
        self.mock_tasks_repo.send_task.assert_called_once()

    def test_send_task_module_with_all_stage_types(self):
        """Test module containing forms, hooks, scripts, and jobs"""
        self.create_module_abstra_json(
            "complex-module",
            stages={
                "forms": [
                    {
                        "id": "mod-form",
                        "path": "modform",
                        "title": "Mod Form",
                        "file": "form.py",
                        "workflow_position": [0, 0],
                        "is_initial": True,
                        "notification_trigger": {
                            "variable_name": "assignee_emails",
                            "enabled": False,
                        },
                        "auto_start": False,
                        "end_message": None,
                        "start_message": None,
                        "error_message": None,
                        "timeout_message": None,
                        "start_button_text": None,
                        "access_control": {"is_public": False, "required_roles": []},
                        "input": True,
                        "output": False,
                    }
                ],
                "hooks": [
                    {
                        "id": "mod-hook",
                        "path": "modhook",
                        "file": "hook.py",
                        "title": "Mod Hook",
                        "enabled": True,
                        "workflow_position": [100, 0],
                        "is_initial": False,
                        "input": False,
                        "output": False,
                    }
                ],
                "scripts": [
                    {
                        "id": "mod-script",
                        "file": "script.py",
                        "title": "Mod Script",
                        "workflow_position": [200, 0],
                        "is_initial": False,
                        "input": False,
                        "output": True,
                    }
                ],
            },
        )

        project = self.repositories.project.load()

        component = ComponentStage(
            id="comp-1",
            title="Complex Module",
            package_name="complex-module",
            workflow_position=(50, 0),
            workflow_transitions=[],
        )

        main_form = FormStage(
            id="main-form",
            path="main",
            title="Main Form",
            file="main.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="trans-1",
                    target_id="comp-1",
                    target_type="components",
                    type="task",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

        project.forms.append(main_form)
        project.components.append(component)
        self.repositories.project.save(project)

        # Send to component - should expand to form input
        self.mock_tasks_repo.send_task = Mock(
            return_value=self.create_mock_task("mod-form")
        )

        self.get_executor().send_task(
            type="task",
            current_stage=main_form,
            payload={"data": "test"},
            execution=None,
        )

        # Should go to module form (input stage)
        call_args = self.mock_tasks_repo.send_task.call_args[1]
        self.assertEqual(call_args["target_stage_id"], "mod-form")



from abstra_internals.interface.sdk.tasks import (
    Task,
    get_sent_tasks,
    get_tasks,
    get_trigger_task,
    iter_tasks,
    send_task,
)

__all__ = [
    "get_tasks",
    "send_task",
    "Task",
    "get_trigger_task",
    "iter_tasks",
    "get_sent_tasks",
]



import unittest
from pathlib import Path

from abstra_internals.utils.file import (
    generate_conflictless_path,
    module2path,
    path2module,
)


class Module2PathTest(unittest.TestCase):
    def test_module2path_module(self):
        module = "a.b.c"
        path = module2path(module, False)
        self.assertEqual(path, Path("a", "b", "c.py"))

    def test_module2path_package(self):
        module = "a.b.c"
        path = module2path(module, True)
        self.assertEqual(path, Path("a", "b", "c", "__init__.py"))


class Path2ModuleTest(unittest.TestCase):
    def test_path2module_module(self):
        path = Path("a", "b", "c.py")
        module = path2module(path)
        self.assertEqual(module, "a.b.c")

    def test_path2module_package(self):
        path = Path("a", "b", "c", "__init__.py")
        module = path2module(path)
        self.assertEqual(module, "a.b.c")


class CreatePathTest(unittest.TestCase):
    def test_without_conflict(self):
        self.assertEqual(generate_conflictless_path("file"), "file")

    def test_conflict_with_static_path(self):
        path = "login"
        generated_path = generate_conflictless_path(path)

        self.assertNotEqual(generated_path, path)
        self.assertIn(path + "-", generated_path)

    def test_conflict_with_dynamic_path(self):
        path = "error/some-error"
        generated_path = generate_conflictless_path(path)

        self.assertNotEqual(generated_path, path)
        self.assertIn(path + "-", generated_path)



from abstra_internals.interface.sdk.tables.api import (
    TablesExecutionError,
    delete,
    delete_by_id,
    insert,
    query_df,
    run,
    run_sql,
    select,
    select_by_id,
    select_df,
    select_one,
    update,
    update_by_id,
)
from abstra_internals.interface.sdk.tables.comparators import (
    is_between,
    is_eq,
    is_gt,
    is_gte,
    is_ilike,
    is_in,
    is_like,
    is_lt,
    is_lte,
    is_neq,
    is_not_ilike,
    is_not_in,
    is_not_like,
    is_not_null,
    is_null,
)

__all__ = [
    "run",
    "run_sql",
    "query_df",
    "TablesExecutionError",
    "select",
    "select_by_id",
    "select_df",
    "select_one",
    "insert",
    "update",
    "delete",
    "update_by_id",
    "delete_by_id",
    "is_eq",
    "is_neq",
    "is_gt",
    "is_between",
    "is_gte",
    "is_in",
    "is_lt",
    "is_like",
    "is_lte",
    "is_not_in",
    "is_not_like",
    "is_null",
    "is_not_null",
    "is_ilike",
    "is_not_ilike",
]



from unittest import TestCase

from abstra_internals.repositories.project.json_migrations.migration_016 import (
    Migration016,
)


class TestMigration016(TestCase):
    def test_removes_deprecated_form_fields(self):
        data = {
            "forms": [
                {
                    "id": "f1",
                    "title": "Form A",
                    "file": "form_a.py",
                    "path": "/form-a",
                    "is_initial": True,
                    "auto_start": False,
                    "allow_restart": True,
                    "restart_button_text": "Start Over",
                    "welcome_title": "Welcome!",
                    "end_message": None,
                    "start_message": None,
                    "error_message": None,
                    "timeout_message": None,
                    "start_button_text": "Begin",
                    "notification_trigger": {
                        "variable_name": "notify",
                        "enabled": False,
                    },
                    "access_control": {"is_public": True, "required_roles": []},
                    "other_field": "should remain",
                }
            ],
            "scripts": [],
            "jobs": [],
            "hooks": [],
            "version": "15.0",
        }
        m = Migration016(data)
        m.apply()

        self.assertEqual(m.data["version"], "16.0")

        form = m.data["forms"][0]
        self.assertNotIn("allow_restart", form)
        self.assertNotIn("restart_button_text", form)
        self.assertNotIn("welcome_title", form)

        self.assertEqual(form["other_field"], "should remain")
        self.assertEqual(form["title"], "Form A")
        self.assertEqual(form["path"], "/form-a")
        self.assertEqual(form["is_initial"], True)
        self.assertEqual(form["auto_start"], False)

    def test_handles_multiple_forms(self):
        data = {
            "forms": [
                {
                    "id": "f1",
                    "title": "Form 1",
                    "file": "form_1.py",
                    "path": "/form-1",
                    "is_initial": True,
                    "auto_start": False,
                    "allow_restart": True,
                    "restart_button_text": "Restart 1",
                    "welcome_title": "Welcome 1",
                    "end_message": None,
                    "start_message": None,
                    "error_message": None,
                    "timeout_message": None,
                    "start_button_text": None,
                    "notification_trigger": {"variable_name": "n", "enabled": False},
                    "access_control": {"is_public": True, "required_roles": []},
                },
                {
                    "id": "f2",
                    "title": "Form 2",
                    "file": "form_2.py",
                    "path": "/form-2",
                    "is_initial": False,
                    "auto_start": True,
                    "allow_restart": False,
                    "restart_button_text": None,
                    "welcome_title": "Hello",
                    "end_message": None,
                    "start_message": None,
                    "error_message": None,
                    "timeout_message": None,
                    "start_button_text": None,
                    "notification_trigger": {"variable_name": "n", "enabled": False},
                    "access_control": {"is_public": False, "required_roles": ["admin"]},
                },
            ],
            "scripts": [],
            "jobs": [],
            "hooks": [],
            "version": "15.0",
        }
        m = Migration016(data)
        m.apply()

        for form in m.data["forms"]:
            self.assertNotIn("allow_restart", form)
            self.assertNotIn("restart_button_text", form)
            self.assertNotIn("welcome_title", form)

            self.assertIn("path", form)
            self.assertIn("is_initial", form)
            self.assertIn("auto_start", form)

    def test_handles_forms_without_deprecated_fields(self):
        data = {
            "forms": [
                {
                    "id": "f1",
                    "title": "Form Without Deprecated Fields",
                    "file": "form.py",
                    "path": "/form",
                    "is_initial": True,
                    "auto_start": False,
                    "end_message": None,
                    "start_message": None,
                    "error_message": None,
                    "timeout_message": None,
                    "start_button_text": None,
                    "notification_trigger": {"variable_name": "n", "enabled": False},
                    "access_control": {"is_public": True, "required_roles": []},
                }
            ],
            "scripts": [],
            "jobs": [],
            "hooks": [],
            "version": "15.0",
        }
        m = Migration016(data)
        m.apply()

        self.assertEqual(m.data["version"], "16.0")
        self.assertEqual(len(m.data["forms"]), 1)



import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.services.requirements import (
    RequirementsRepository,
    create_requirement,
    uninstall_requirement,
)
from abstra_internals.usage import editor_usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_requirements", __name__)

    # 1s pooling in this route
    @bp.get("/")
    def _get_requirements():
        return RequirementsRepository.load().to_dict()

    @bp.post("/")
    @editor_usage
    def _create_requirement():
        data = flask.request.json

        if not data:
            flask.abort(400)

        name = data["name"]
        version = data.get("version")

        requirements = RequirementsRepository.load()
        requirements.add(name, version)
        RequirementsRepository.save(requirements)
        return requirements.to_dict()

    @bp.post("/install")
    def _install_requirements():
        requirements = RequirementsRepository.load()
        streamer = requirements.install()
        if streamer is None:
            flask.abort(403)

        return flask.Response(streamer, mimetype="text/event-stream")

    @bp.post("/<name>/uninstall")
    def _uninstall_requirement(name: str):
        req = create_requirement(name)
        streamer = uninstall_requirement(req)
        if streamer is None:
            flask.abort(403)
        reqs = RequirementsRepository.load()
        reqs.delete(name)
        RequirementsRepository.save(reqs)
        return flask.Response(streamer, mimetype="text/event-stream")

    @bp.delete("/<name>")
    @editor_usage
    def _delete_requirement(name: str):
        requirements = RequirementsRepository.load()
        requirements.delete(name)
        RequirementsRepository.save(requirements)
        return requirements.to_dict()

    # 1s pooling in this route
    @bp.get("/recommendations")
    def _get_requirements_recommendation():
        return [r.to_dict() for r in RequirementsRepository.get_recommendation()]

    return bp
